# Functions to read and access the zeekctl configuration.

import configparser
import hashlib
import os
import re
import socket
import subprocess
import sys

from ZeekControl import events, options
from ZeekControl import node as node_mod
from ZeekControl.exceptions import ConfigurationError, RuntimeEnvironmentError

from .state import SqliteState
from .version import VERSION

# Class storing the zeekctl configuration.
#
# This class provides access to four types of configuration/state:
#
# - the global zeekctl configuration from zeekctl.cfg
# - the node configuration from node.cfg
# - dynamic state variables which are kept across restarts in spool/state.db

Config = None  # Globally accessible instance of Configuration.


class NodeStore:
    def __init__(self):
        self.nodestore = {}
        self.nodenameslower = []

    def add_node(self, node):
        # Add a node to the nodestore, but first check for duplicate node
        # names. This check is not case-sensitive, because names are stored
        # lowercase in the state db, and some filesystems are not
        # case-sensitive (working dir name is node name).
        # Duplicate node names can occur either because the user defined two
        # nodes that differ only by case (e.g. "Worker-1" and "worker-1"), or
        # if a user defines a node name that conflicts with an auto-generated
        # one (e.g. "worker-1" with lb_procs=2 and "worker-1-2").
        namelower = node.name.lower()
        if namelower in self.nodenameslower:
            matchname = ""
            for nn in self.nodestore:
                if nn.lower() == namelower:
                    matchname = nn
                    break
            raise ConfigurationError(
                f'node name "{node.name}" is a duplicate of "{matchname}"'
            )

        self.nodestore[node.name] = node
        self.nodenameslower.append(namelower)


class Configuration:
    def __init__(
        self, basedir, libdir, libdirinternal, cfgfile, zeekscriptdir, ui, state=None
    ):
        self.ui = ui
        self.basedir = basedir
        self.libdir = libdir
        self.libdirinternal = libdirinternal
        self.cfgfile = cfgfile
        self.zeekscriptdir = zeekscriptdir
        global Config
        Config = self

        self.config = {}
        self.state = {}
        self.nodestore = {}

        self.localaddrs = self._get_local_addrs()

        # Read zeekctl.cfg.
        self.config = self._read_config(cfgfile)

        self._initialize_options()
        self._check_options()

        if state:
            self.state_store = state
        else:
            self.state_store = SqliteState(self.statefile)

        self.read_state()
        self._update_cfg_state()

    def reload_cfg(self):
        self.config = self._read_config(self.cfgfile)
        self._initialize_options()
        self._check_options()
        self._update_cfg_state()

    def _initialize_options(self):
        from ZeekControl import execute

        # Set defaults for options we get passed in.
        self.init_option("zeekbase", self.basedir)
        self.init_option("zeekscriptdir", self.zeekscriptdir)
        self.init_option("version", VERSION)
        self.init_option("libdir", self.libdir)
        self.init_option("libdirinternal", self.libdirinternal)

        # Initialize options that are not already set.
        errors = False
        for opt in options.options:
            if opt.dontinit:
                continue

            if opt.legacy_name:
                old_key = opt.legacy_name.lower()
                if old_key in self.config:
                    self.ui.error(
                        f"option '{opt.legacy_name}' is no longer supported, please use '{opt.name}' instead"
                    )
                    errors = True
                    continue

            self.init_option(opt.name, opt.default)

        if errors:
            sys.exit(1)

        # Set defaults for options we derive dynamically.
        self.init_option("mailto", "{}".format(os.getenv("USER")))
        self.init_option("mailfrom", f"Zeek <zeek@{socket.gethostname()}>")
        self.init_option("mailalarmsto", self.config["mailto"])

        # Determine operating system.
        success, output = execute.run_localcmd("uname")
        if not success or not output:
            raise RuntimeEnvironmentError(f"failed to run uname: {output}")
        self.init_option("os", output.strip())

        # Determine the CPU pinning command.
        pin_cmd = ""
        if self.config["os"] == "Linux":
            pin_cmd = "taskset -c"
        elif self.config["os"] == "FreeBSD":
            pin_cmd = "cpuset -l"

        self.init_option("pin_command", pin_cmd)

        # Find the time command (should be a GNU time for best results).
        time_cmd = ""
        success, output = execute.run_localcmd("which time")
        if success and output:
            # On redhat-based systems, path to cmd is prefixed with '\t' on 2nd
            # line when alias is defined.
            time_cmd = output.splitlines()[-1].strip()

        self.init_option("time", time_cmd)

        # Calculate the log expire interval (in minutes).
        minutes = self._get_interval_minutes("logexpireinterval")
        self.init_option("logexpireminutes", minutes)

    # Do a basic sanity check on zeekctl options.
    def _check_options(self):
        # Option names must be valid bash variable names because we will
        # write them to zeekctl-config.sh (note that zeekctl will convert "."
        # to "_" when it writes to zeekctl-config.sh).
        allowedchars = re.compile("^[a-z0-9_.]+$")
        nostartdigit = re.compile("^[^0-9]")

        for key, value in self.config.items():
            if re.match(allowedchars, key) is None:
                raise ConfigurationError(
                    f'zeekctl option name "{key}" contains invalid characters (allowed characters: a-z, 0-9, ., and _)'
                )
            if re.match(nostartdigit, key) is None:
                raise ConfigurationError(
                    f'zeekctl option name "{key}" cannot start with a number'
                )

            # No zeekctl option ever requires the entire value to be wrapped in
            # quotes, and since doing so can cause problems, we don't allow it.
            if isinstance(value, str):
                if (
                    value.startswith('"')
                    and value.endswith('"')
                    or value.startswith("'")
                    and value.endswith("'")
                ):
                    raise ConfigurationError(
                        f'value of zeekctl option "{key}" cannot be wrapped in quotes'
                    )

        dirs = (
            "zeekbase",
            "logdir",
            "spooldir",
            "cfgdir",
            "zeekscriptdir",
            "bindir",
            "libdirinternal",
            "plugindir",
            "scriptsdir",
        )
        files = ("makearchivename",)

        for d in dirs:
            v = self.config[d]
            if not os.path.isdir(v):
                raise ConfigurationError(
                    f'zeekctl option "{d}" directory not found: {v}'
                )

        for f in files:
            v = self.config[f]
            if not os.path.isfile(v):
                raise ConfigurationError(f'zeekctl option "{f}" file not found: {v}')

        # Verify that logs don't expire more quickly than the rotation interval
        logexpireseconds = 60 * self.config["logexpireminutes"]
        if 0 < logexpireseconds < self.config["logrotationinterval"]:
            raise ConfigurationError(
                "Log expire interval cannot be shorter than the log rotation interval"
            )

        if self.config["usewebsocket"]:
            if events.websockets_errmsg is not None:
                self.ui.warn(
                    f"zeekctl option UseWebSocket is set, but websockets non-functional: {events.websockets_errmsg}"
                )
                self.ui.warn(
                    "The print and netstats commands will be non-functional, please install the Python websockets package version 11.0 or later."
                )

        if not self.config["usewebsocket"]:
            # If UseWebSocket isn't configured, then commands for non-Broker
            # backends will not work. Warn the user here. We could imply UseWebSocket = 1,
            # but that results in the manager listening on a new port, so better
            # to not have the user explicitly toggle it on.
            clusterbackend = self.config["clusterbackend"]
            if clusterbackend.lower() != "broker":
                self.ui.warn(
                    f"zeekctl netstats and print commands with cluster backend '{clusterbackend}' require UseWebSocket = 1"
                )

    # Convert a time interval string (from the value of the given option name)
    # to an integer number of minutes.
    def _get_interval_minutes(self, optname):
        # Conversion table for time units to minutes.
        units = {"day": 24 * 60, "hr": 60, "min": 1}

        ss = self.config[optname]
        try:
            # If no time unit, assume it's days (for backward compatibility).
            v = int(ss) * units["day"]
            return v
        except ValueError:
            pass

        # Time interval is a non-negative integer followed by an optional
        # space, followed by a time unit.
        mm = re.match("([0-9]+) ?(day|hr|min)s?$", ss)
        if mm is None:
            raise ConfigurationError(
                f'value of zeekctl option "{optname}" is invalid (value must be integer followed by a time unit "day", "hr", or "min"): {ss}'
            )

        v = int(mm.group(1))
        v *= units[mm.group(2)]

        return v

    def initPostPlugins(self):
        # Read node.cfg
        self.nodestore = self._read_nodes()

        # If "env_vars" was specified in zeekctl.cfg, then apply to all nodes.
        varlist = self.config.get("env_vars")
        if varlist:
            try:
                global_env_vars = self._get_env_var_dict(varlist)
            except ConfigurationError as err:
                raise ConfigurationError(f"zeekctl config: {err}")

            for node in self.nodes():
                for key, val in global_env_vars.items():
                    # Values from node.cfg take precedence over zeekctl.cfg
                    node.env_vars.setdefault(key, val)

        # Check state store for any running nodes that are no longer in the
        # current node config.
        self._warn_dangling_zeek()

        # Set the standalone config option.
        standalone = len(self.nodestore) == 1

        self.init_option("standalone", standalone)

    # Provides access to the configuration options via the dereference operator.
    # Lookup the attribute in zeekctl options first, then in the dynamic state
    # variables.
    def __getattr__(self, attr):
        if attr in self.config:
            return self.config[attr]
        if attr in self.state:
            return self.state[attr]
        raise AttributeError(f"unknown config attribute {attr}")

    # Returns a sorted list of all zeekctl.cfg entries.
    # Includes dynamic variables if dynamic is true.
    def options(self, dynamic=True):
        optlist = list(self.config.items())
        if dynamic:
            optlist += list(self.state.items())

        optlist.sort()
        return optlist

    # Returns a list of Nodes (the list will be empty if no matching nodes
    # are found).  The returned list is sorted by node type, and by node name
    # for each type.
    # - By default (i.e. tag is None), all Nodes are returned.
    # - If tag is a node group name (e.g. "workers"), all nodes belonging to
    #   that group are returned.
    # - If tag is the name of a node, then that node is returned.
    def nodes(self, tag=None):
        nodetype = node_mod.group_type(tag)
        if nodetype == "_ALL_":
            tag = None

        nodes = []
        for n in self.nodestore.values():
            if nodetype == n.type or tag == n.name or tag is None:
                nodes += [n]

        nodes.sort(key=node_mod.sortnode)

        return nodes

    # Returns the manager Node (cluster config) or standalone Node (standalone
    # config).  Returns None if neither are available.
    def manager(self):
        if self.config["standalone"]:
            n = self.nodes()
        else:
            n = self.nodes(node_mod.manager_group())

        if not n:
            return None

        return n[0]

    def loggers(self):
        return self.nodes(node_mod.logger_group())

    def proxies(self):
        return self.nodes(node_mod.proxy_group())

    def workers(self):
        return self.nodes(node_mod.worker_group())

    # Returns a list of nodes which is a subset of the result a similar call to
    # nodes() would yield but within which each host appears only once.
    # If "exclude_local" is True, then the returned list will not include
    # nodes that are on the local host.
    def hosts(self, tag=None, exclude_local=False):
        hosts = {}
        nodelist = []
        for node in self.nodes(tag):
            if node.host in hosts:
                continue

            if exclude_local and node.addr in self.localaddrs:
                continue

            hosts[node.host] = 1
            nodelist.append(node)

        return nodelist

    # Replace all occurences of "${option}", with option being either
    # zeekctl.cfg option or a dynamic variable, with the corresponding value.
    # Defaults to replacement with the empty string for unknown options.
    def subst(self, text):
        while True:
            match = re.search(r"(\$\{([A-Za-z][A-Za-z0-9]*)(:([^}]+))?\})", text)
            if not match:
                return text

            key = match.group(2).lower()
            try:
                value = str(self.__getattr__(key))
            except AttributeError:
                value = match.group(4)
                if value is None:
                    value = ""

            text = text[0 : match.start(1)] + value + text[match.end(1) :]

    # Convert string into list of integers (ValueError is raised if any
    # item in the list is not a non-negative integer).
    def _get_pin_cpu_list(self, text, numprocs):
        if not text:
            return []

        cpulist = [int(x) for x in text.split(",")]
        # Minimum allowed CPU number is zero.
        if min(cpulist) < 0:
            raise ValueError

        # Make sure list is at least as long as number of worker processes.
        cpulen = len(cpulist)
        if numprocs > cpulen:
            cpulist = [cpulist[i % cpulen] for i in range(numprocs)]

        return cpulist

    # Convert a string consisting of a comma-separated list of environment
    # variables (e.g. "VAR1=123, VAR2=456") to a dictionary.
    # If the string is empty, then return an empty dictionary.
    def _get_env_var_dict(self, text):
        env_vars = {}

        if text:
            for keyval in text.split(","):
                try:
                    key, val = keyval.split("=", 1)
                except ValueError:
                    raise ConfigurationError(
                        f"missing '=' in env_vars option value: {keyval}"
                    )

                key = key.strip()
                if not key:
                    raise ConfigurationError(
                        f"env_vars option value must contain at least one environment variable name: {keyval}"
                    )

                env_vars[key] = val.strip()

        return env_vars

    # Parse node.cfg.
    def _read_nodes(self):
        config = configparser.ConfigParser()
        fname = self.nodecfg
        try:
            if not config.read(fname):
                raise ConfigurationError(f"cannot read node config file: {fname}")
        except configparser.MissingSectionHeaderError as err:
            raise ConfigurationError(err)

        nodestore = NodeStore()

        counts = {}
        for sec in config.sections():
            node = node_mod.Node(self, sec)

            # Note that the keys are converted to lowercase by configparser.
            for key, val in config.items(sec):
                key = key.replace(".", "_")

                if key not in node_mod.Node._keys:
                    self.ui.warn(
                        f"ignoring unrecognized node config option '{key}' given for node '{sec}'"
                    )
                    continue

                node.__dict__[key] = val

            # Perform a sanity check on the node, and update nodestore.
            self._check_node(node, nodestore, counts)

        # Perform a sanity check on the nodestore (make sure we have a valid
        # cluster config, etc.).
        self._check_nodestore(nodestore.nodestore)

        return nodestore.nodestore

    def _check_node(self, node, nodestore, counts):
        if not node.type:
            raise ConfigurationError(f"no type given for node {node.name}")

        if node.type not in node_mod.node_types():
            raise ConfigurationError(
                f"unknown node type '{node.type}' given for node '{node.name}'"
            )

        if not node.host:
            raise ConfigurationError(f"no host given for node '{node.name}'")

        try:
            addrinfo = socket.getaddrinfo(node.host, None, 0, 0, socket.SOL_TCP)
        except socket.gaierror as e:
            raise ConfigurationError(
                f"hostname lookup failed for '{node.host}' in node config [{e.args[1]}]"
            )

        addrs = [addr[4][0] for addr in addrinfo]

        # By default, just use the first IP addr in the list.
        addr_str = addrs[0]

        # Choose the first IPv4 addr (if any) in the list.
        for ip in addrs:
            if ":" not in ip:
                addr_str = ip
                break

        # zone_id is handled manually, so strip it if it's there
        node.addr = addr_str.split("%")[0]

        # Convert env_vars from a string to a dictionary.
        try:
            node.env_vars = self._get_env_var_dict(node.env_vars)
        except ConfigurationError as err:
            raise ConfigurationError(f"node '{node.name}' config: {err}")

        # Each node gets a number unique across its type.
        try:
            counts[node.type] += 1
        except KeyError:
            counts[node.type] = 1

        node.count = counts[node.type]

        numprocs = 0

        if node.lb_procs:
            if not node_mod.is_worker(node):
                raise ConfigurationError(
                    f"node '{node.name}' config: load balancing node config options are only for worker nodes"
                )
            try:
                numprocs = int(node.lb_procs)
            except ValueError:
                raise ConfigurationError(
                    f"number of load-balanced processes must be an integer for node '{node.name}'"
                )
            if numprocs < 1:
                raise ConfigurationError(
                    f"number of load-balanced processes must be greater than zero for node '{node.name}'"
                )
        elif node.lb_method:
            raise ConfigurationError(
                f"number of load-balanced processes not specified for node '{node.name}'"
            )

        try:
            pin_cpus = self._get_pin_cpu_list(node.pin_cpus, numprocs)
        except ValueError:
            raise ConfigurationError(
                f"pin cpus list must contain only non-negative integers for node '{node.name}'"
            )

        if pin_cpus:
            node.pin_cpus = pin_cpus[0]

        if node.lb_procs:
            if not node.lb_method:
                raise ConfigurationError(
                    f"no load balancing method given for node '{node.name}'"
                )

            if node.lb_method not in (
                "af_packet",
                "pf_ring",
                "myricom",
                "custom",
                "interfaces",
            ):
                raise ConfigurationError(
                    f"unknown load balancing method '{node.lb_method}' given for node '{node.name}'"
                )

            if node.lb_method == "interfaces":
                if not node.lb_interfaces:
                    raise ConfigurationError(
                        f"list of load-balanced interfaces not specified for node '{node.name}'"
                    )

                # get list of interfaces to use, and assign one to each node
                netifs = node.lb_interfaces.split(",")

                if len(netifs) != numprocs:
                    raise ConfigurationError(
                        f"number of load-balanced interfaces is not same as number of load-balanced processes for node '{node.name}'"
                    )

                node.interface = netifs.pop().strip()

            origname = node.name
            # node names will have a numerical suffix
            node.name = f"{node.name}-1"

            for num in range(2, numprocs + 1):
                newnode = node.copy()

                # Update the node attrs that need to be changed
                newname = f"{origname}-{num:d}"
                newnode.name = newname
                counts[node.type] += 1
                newnode.count = counts[node.type]
                if pin_cpus:
                    newnode.pin_cpus = pin_cpus[num - 1]

                if newnode.lb_method == "interfaces":
                    newnode.interface = netifs.pop().strip()

                nodestore.add_node(newnode)

        nodestore.add_node(node)

    def _check_nodestore(self, nodestore):
        if not nodestore:
            raise ConfigurationError("no nodes found in node config")

        standalone = False
        manager = False
        proxy = False

        manageronlocalhost = False
        # Note: this is a subset of localaddrs
        localhostaddrs = "127.0.0.1", "::1"

        for n in nodestore.values():
            if node_mod.is_manager(n):
                if manager:
                    raise ConfigurationError(
                        "only one manager can be defined in node config"
                    )
                manager = True
                if n.addr in localhostaddrs:
                    manageronlocalhost = True

                if n.addr not in self.localaddrs:
                    raise ConfigurationError(
                        "must run zeekctl on same machine as the manager node. The manager node has IP address {} and this machine has IP addresses: {}".format(
                            n.addr, ", ".join(self.localaddrs)
                        )
                    )

            elif node_mod.is_proxy(n):
                proxy = True

            elif node_mod.is_standalone(n):
                standalone = True
                if n.addr not in self.localaddrs:
                    raise ConfigurationError(
                        "must run zeekctl on same machine as the standalone node. The standalone node has IP address {} and this machine has IP addresses: {}".format(
                            n.addr, ", ".join(self.localaddrs)
                        )
                    )

        if standalone:
            if len(nodestore) > 1:
                raise ConfigurationError(
                    "more than one node defined in standalone node config"
                )
        else:
            if not manager:
                raise ConfigurationError("no manager defined in node config")
            elif not proxy:
                raise ConfigurationError("no proxy defined in node config")

        # If manager is on localhost, then all other nodes must be on localhost
        if manageronlocalhost:
            for n in nodestore.values():
                if not node_mod.is_manager(n) and n.addr not in localhostaddrs:
                    raise ConfigurationError(
                        "all nodes must use localhost/127.0.0.1/::1 when manager uses it"
                    )

    def _to_bool(self, val):
        if val.lower() in ("1", "true"):
            return True
        if val.lower() in ("0", "false"):
            return False
        raise ValueError(f"invalid boolean: '{val}'")

    # Parses zeekctl.cfg and returns a dictionary of all entries.
    def _read_config(self, fname):
        type_converters = {"bool": self._to_bool, "int": int, "string": str}
        config = {}

        opt_names = set()
        for opt in options.options:
            # Convert key to lowercase because keys are stored in lowercase.
            key = opt.name.lower()
            opt_names.add(key)
            if opt.legacy_name:
                opt_names.add(opt.legacy_name.lower())

        with open(fname) as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith("#"):
                    continue

                args = line.split("=", 1)
                if len(args) != 2:
                    raise ConfigurationError(f"zeekctl config syntax error: {line}")

                key, val = args
                # Option names are not case-sensitive.
                key = key.strip().lower()

                # Warn about unrecognized options, but we can't check plugin
                # options here because no plugins have been loaded yet.
                if "." not in key and key not in opt_names:
                    self.ui.warn(f"ignoring unrecognized zeekctl option: {key}")
                    continue

                # if the key already exists, just overwrite with new value
                config[key] = val.strip()

        # Convert option values to correct data type
        for opt in options.options:
            # Convert key to lowercase because keys are stored in lowercase.
            key = opt.name.lower()
            if key in config:
                try:
                    config[key] = type_converters[opt.type](config[key])
                except ValueError:
                    raise ConfigurationError(
                        f"zeekctl option '{key}' has invalid value '{config[key]}' for type {opt.type}"
                    )

        return config

    # Initialize a global option if not already set.
    def init_option(self, key, val):
        # Store option names in lowercase, because they are not case-sensitive.
        key = key.lower()

        if key not in self.config:
            if isinstance(val, str):
                self.config[key] = self.subst(val)
            else:
                self.config[key] = val

    # Set a global option (regardless of whether or not it is already set).
    def set_option(self, key, val):
        # Store option names in lowercase, because they are not case-sensitive.
        key = key.lower()

        self.config[key] = val

    # Returns value of an option, or None if the option is not defined.
    def get_option(self, key):
        # Convert key to lowercase because keys are stored in lowercase.
        return self.config.get(key.lower())

    # Set a dynamic state variable.
    def set_state(self, key, val):
        key = key.lower()
        if self.state.get(key) == val:
            return

        self.state[key] = val
        self.state_store.set(key, val)

    # Returns value of state variable, or the specified default value if the
    # state variable is not defined.
    def get_state(self, key, default=None):
        return self.state.get(key.lower(), default)

    # Read dynamic state variables.
    def read_state(self):
        self.state = dict(self.state_store.items())

    # Use the ifconfig command to find local IP addrs.
    def _get_local_addrs_ifconfig(self):
        try:
            # On Linux, ifconfig is often not in the user's standard PATH.
            # Also need to set LANG here to ensure that the output of ifconfig
            # is consistent regardless of which locale the system is using.
            proc = subprocess.Popen(
                ["PATH=$PATH:/sbin:/usr/sbin LANG=C ifconfig", "-a"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=True,
            )
            out, err = proc.communicate()
        except OSError:
            return False

        success = proc.returncode == 0
        if not success:
            return False

        localaddrs = []
        out = out.decode()

        # The output of ifconfig varies by OS and by IPv4 vs IPv6.
        # Linux example:
        #   inet addr:127.0.0.1
        #   inet6 addr: ::1/128
        # BSD (and OS X) example:
        #   inet 127.0.0.1
        #   inet6 ::1
        #   inet6 fe80::1%lo0
        for line in out.splitlines():
            fields = line.split()
            if len(fields) < 3:
                continue

            if fields[0] != "inet" and fields[0] != "inet6":
                continue

            addrstr = fields[1]

            if addrstr[-1] == ":" and addrstr.count(":") == 1:
                addrstr = fields[2]

            if addrstr.count(":") == 1:
                # Remove "addr:" prefix (if any).
                addrstr = addrstr.split(":")[1]

            # Remove everything after "/" or "%" (if any)
            addrstr = addrstr.split("/")[0]
            addrstr = addrstr.split("%")[0]

            if _is_valid_addr(addrstr):
                localaddrs.append(addrstr)

        if not localaddrs:
            self.ui.warn(
                'failed to extract IP addresses from the "ifconfig -a" command output'
            )

        return localaddrs

    # Use the ip command to find local IP addrs.
    def _get_local_addrs_ip(self):
        try:
            # On Linux, "ip" is sometimes not in the user's standard PATH.
            proc = subprocess.Popen(
                ["PATH=$PATH:/sbin:/usr/sbin ip address"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=True,
            )
            out, err = proc.communicate()
        except OSError:
            return False

        success = proc.returncode == 0
        if not success:
            return False

        localaddrs = []
        out = out.decode()

        # Here is an example portion of "ip" command output:
        #    inet 127.0.0.1/8
        #    inet6 ::1/128
        for line in out.splitlines():
            fields = line.split()
            if len(fields) < 2:
                continue

            if fields[0] != "inet" and fields[0] != "inet6":
                continue

            addrstr = fields[1]
            addrstr = addrstr.split("/")[0]
            addrstr = addrstr.split("%")[0]

            if _is_valid_addr(addrstr):
                localaddrs.append(addrstr)

        if not localaddrs:
            self.ui.warn(
                'failed to extract IP addresses from the "ip address" command output'
            )

        return localaddrs

    # Return a list of the IP addresses associated with local interfaces.
    # For IPv6 addresses, zone_id and prefix length are removed if present.
    def _get_local_addrs(self):
        # ifconfig is more portable so try it first
        localaddrs = self._get_local_addrs_ifconfig()

        if not localaddrs:
            # On some Linux systems ifconfig has been superseded by ip.
            localaddrs = self._get_local_addrs_ip()

        # Fallback to localhost if we did not find any IP addrs.
        if not localaddrs:
            self.ui.warn(
                'failed to find local IP addresses with "ifconfig -a" or "ip address" commands'
            )

            localaddrs = ["127.0.0.1", "::1"]
            try:
                addrinfo = socket.getaddrinfo(
                    socket.gethostname(), None, 0, 0, socket.SOL_TCP
                )
            except Exception:
                addrinfo = []

            for ai in addrinfo:
                localaddrs.append(ai[4][0])

        return localaddrs

    # Record the Zeek version.
    def record_zeek_version(self):
        version = self._get_zeek_version()
        self.set_state("zeekversion", version)

    # Record the state of the zeekctl config files.
    def _update_cfg_state(self):
        self.set_state("configchksum", self._get_zeekctlcfg_hash(filehash=True))
        self.set_state("confignodechksum", self._get_nodecfg_hash(filehash=True))

    # Returns True if the zeekctl config files have changed since last reload.
    def is_cfg_changed(self):
        try:
            if "configchksum" in self.state:
                if self.state["configchksum"] != self._get_zeekctlcfg_hash(
                    filehash=True
                ):
                    return True

            if "confignodechksum" in self.state:
                if self.state["confignodechksum"] != self._get_nodecfg_hash(
                    filehash=True
                ):
                    return True
        except OSError:
            # If we can't read the config files, then do nothing.
            pass

        return False

    # Check if the user has already run the "install" or "deploy" commands.
    def is_zeekctl_installed(self):
        return os.path.isfile(
            os.path.join(self.config["policydirsiteinstallauto"], "zeekctl-config.zeek")
        )

    # Warn user to run zeekctl deploy if any changes are detected to zeekctl
    # config options, node config, Zeek version, or if certain state variables
    # are missing.
    def warn_zeekctl_install(self):
        missingstate = False

        # Check if node config has changed since last install.
        if "hash-nodecfg" in self.state:
            nodehash = self._get_nodecfg_hash()

            if nodehash != self.state["hash-nodecfg"]:
                self.ui.warn(
                    'zeekctl node config has changed (run the zeekctl "deploy" command)'
                )

                return
        else:
            missingstate = True

        # If this is a fresh install (i.e., zeekctl install not yet run), then
        # inform the user what to do.
        if not self.is_zeekctl_installed():
            self.ui.info('Hint: Run the zeekctl "deploy" command to get started.')
            return

        # Check if Zeek version is different from the previously-installed
        # version.
        if "zeekversion" in self.state:
            oldversion = self.state["zeekversion"]

            version = self._get_zeek_version()

            if version != oldversion:
                self.ui.warn(
                    'new zeek version detected (run the zeekctl "deploy" command)'
                )
                return
        else:
            missingstate = True

        # Check if any config values have changed since last install.
        if "hash-zeekctlcfg" in self.state:
            cfghash = self._get_zeekctlcfg_hash()
            if cfghash != self.state["hash-zeekctlcfg"]:
                self.ui.warn(
                    'zeekctl config has changed (run the zeekctl "deploy" command)'
                )
                return
        else:
            missingstate = True

        # If any of the state variables don't exist, then we need to install
        # (this would most likely indicate an upgrade install was performed
        # over an old version that didn't have the state.db file).
        if missingstate:
            self.ui.warn(
                'state database needs updating (run the zeekctl "deploy" command)'
            )
            return

    # Warn if there might be any dangling Zeek nodes (i.e., nodes that are
    # still running but are either no longer part of the current node
    # configuration or have moved to a new host).
    def _warn_dangling_zeek(self):
        nodes = {}
        for n in self.nodes():
            # Convert node name to lowercase because below we are using
            # node names from the state db (which is lowercase).
            nodes[n.name.lower()] = n.host

        for key in self.state.keys():
            # Look for a PID associated with a Zeek node
            if not key.endswith("-pid"):
                continue

            pid = self.get_state(key)
            if not pid:
                continue

            # Get node name and host name for this node
            nname = key[:-4]
            hostkey = key.replace("-pid", "-host")
            hname = self.get_state(hostkey)
            if not hname:
                continue

            # If node is not a known node or if host has changed, then
            # we must warn about dangling Zeek node.
            if nname not in nodes or hname != nodes[nname]:
                self.ui.warn(
                    f'Zeek node "{nname}" possibly still running on host "{hname}" (PID {pid})'
                )
                # Set the "expected running" flag to False so cron doesn't try
                # to start this node.
                expectkey = key.replace("-pid", "-expect-running")
                self.set_state(expectkey, False)
                # Clear the PID so we don't keep getting warnings.
                self.set_state(key, None)

    # Return a hash value (as a string) of the current zeekctl configuration.
    def _get_zeekctlcfg_hash(self, filehash=False):
        if filehash:
            with open(self.cfgfile) as ff:
                data = ff.read()
        else:
            data = str(sorted(self.config.items()))

        data = data.encode()

        hh = hashlib.sha1()
        hh.update(data)
        return hh.hexdigest()

    # Return a hash value (as a string) of the current zeekctl node config.
    def _get_nodecfg_hash(self, filehash=False):
        if filehash:
            with open(self.nodecfg) as ff:
                data = ff.read()
        else:
            nn = []
            for n in self.nodes():
                nn.append(
                    tuple(
                        [
                            (key, val)
                            for key, val in n.items()
                            if not key.startswith("_")
                        ]
                    )
                )
            data = str(nn)

        data = data.encode()

        hh = hashlib.sha1()
        hh.update(data)
        return hh.hexdigest()

    # Update the stored hash value of the current zeekctl config.
    def update_cfg_hash(self):
        cfghash = self._get_zeekctlcfg_hash()
        nodehash = self._get_nodecfg_hash()

        self.set_state("hash-zeekctlcfg", cfghash)
        self.set_state("hash-nodecfg", nodehash)

    # Runs Zeek to get its version number.
    def _get_zeek_version(self):
        from ZeekControl import execute

        zeek = self.config["zeek"]
        if not os.path.lexists(zeek):
            raise ConfigurationError(f"cannot find Zeek binary: {zeek}")

        version = ""
        success, output = execute.run_localcmd(f"{zeek} -v")
        if success and output:
            version = output.splitlines()[-1]
        else:
            msg = " with no output"
            if output:
                msg = f" with output:\n{output}"
            raise RuntimeEnvironmentError(f'running "zeek -v" failed{msg}')

        match = re.search(".* version ([^ ]*).*$", version)
        if not match:
            raise RuntimeEnvironmentError(
                f'cannot determine Zeek version ("zeek -v" output: {version.strip()})'
            )

        version = match.group(1)
        # If zeek is built with the "--enable-debug" configure option, then it
        # appends "-debug" to the version string.
        if version.endswith("-debug"):
            version = version[:-6]

        return version


# Check if a string is a valid representation of an IP address or not.
def _is_valid_addr(ipstr):
    try:
        if ":" in ipstr:
            socket.inet_pton(socket.AF_INET6, ipstr)
        else:
            socket.inet_pton(socket.AF_INET, ipstr)
    except OSError:
        return False

    return True



import json
import sqlite3

from ZeekControl.exceptions import RuntimeEnvironmentError


class SqliteState:
    def __init__(self, path):
        self.path = path

        try:
            self.db = sqlite3.connect(self.path)
        except sqlite3.Error as err:
            raise RuntimeEnvironmentError(
                f"{err}: {path}\nCheck if the user running ZeekControl has both write and search permission to\nthe directory containing the database file and has both read and write\npermission to the database file itself."
            )

        self.c = self.db.cursor()

        try:
            self.setup()
        except sqlite3.Error as err:
            raise RuntimeEnvironmentError(
                f"{err}: {path}\nCheck if the user running ZeekControl has write access to the database file.\nOtherwise, the database file is possibly corrupt."
            )

    def setup(self):
        # Create table
        self.c.execute("""CREATE TABLE IF NOT EXISTS state (
            key   TEXT  PRIMARY KEY  NOT NULL,
            value TEXT
        )""")

        self.db.commit()

    def get(self, key):
        self.c.execute("SELECT value FROM state WHERE key=?", [key])
        records = self.c.fetchall()
        if records:
            return json.loads(records[0][0])
        return None

    def set(self, key, value):
        value = json.dumps(value)
        try:
            self.c.execute("REPLACE INTO state (key, value) VALUES (?,?)", [key, value])
        except sqlite3.Error as err:
            raise RuntimeEnvironmentError(
                f"{err}: {self.path}\nCheck if the user running ZeekControl has write access to the database file."
            )

        self.db.commit()

    def items(self):
        self.c.execute("SELECT key, value FROM state")
        return [(k, json.loads(v)) for (k, v) in self.c.fetchall()]



import ZeekControl.plugin


class AF_Packet(ZeekControl.plugin.Plugin):
    def __init__(self):
        super().__init__(apiversion=1)

    def name(self):
        return "af_packet"

    def pluginVersion(self):
        return 1

    def init(self):
        """
        Only use the plugin if there is a worker using AF_PACKET for load balancing.

        There are two ways to recognize AF_PACKET usage:
          * The specified interface already starts with af_packet:: (lb_custom
            or interfaces usage).
          * lb_method is set to "af_packet" in which case the interface
            is prefixed with af_packet:: by this plugin.
        """
        result = False
        for nn in self.nodes():
            if nn.type != "worker" or not nn.lb_procs:
                continue

            if nn.interface.startswith("af_packet::"):
                result = True

            if nn.lb_method == "af_packet":
                if nn.interface.startswith("af_packet::"):
                    self.error(
                        f"unexpected af_packet:: prefix for interface {nn.interface} of {nn}"
                    )
                    return False

                nn.interface = f"af_packet::{nn.interface}"
                result = True

        return result

    def nodeKeys(self):
        return [
            "block_size",
            "block_timeout",
            "buffer_size",
            "checksum_validation_mode",
            "enable_defrag",
            "enable_fanout",
            "enable_hw_timestamping",
            "fanout_id",
            "fanout_mode",
            "link_type",
        ]

    def zeekctl_config(self):
        script = ""

        # Add custom configuration values per worker.
        for nn in self.nodes():
            if nn.type != "worker" or not nn.lb_procs:
                continue

            params = ""

            for key in self.nodeKeys():
                prefixed_key = f"{self.prefix()}_{key}"
                v = getattr(nn, prefixed_key, None)
                if v and v.strip():
                    params += f"\n  redef AF_Packet::{key} = {v};"

            if params:
                script += f'\n@if( peer_description == "{nn.name}" ) {params}\n@endif'

        return script



"""This module provides command line parsers and corresponding commands."""

import argparse
import configparser
import ipaddress
import json
import os
import sys
import traceback
from typing import Any

from . import brokertypes as bt
from .config import CONFIG
from .consts import CONFIG_FILE
from .controller import Controller
from .controller import Error as ControllerError
from .events import (
    DeployRequest,
    DeployResponse,
    GetConfigurationRequest,
    GetConfigurationResponse,
    GetIdValueRequest,
    GetIdValueResponse,
    GetInstancesRequest,
    GetInstancesResponse,
    GetNodesRequest,
    GetNodesResponse,
    RestartRequest,
    RestartResponse,
    StageConfigurationRequest,
    StageConfigurationResponse,
    TestTimeoutRequest,
    TestTimeoutResponse,
)
from .logs import LOG
from .types import (
    ClusterRole,
    Configuration,
    Enum,
    Instance,
    ManagementRole,
    NodeOutputs,
    NodeStatus,
    Result,
)

# For unit-testing, a central place to adjust where reads from stdin come from
# and writes to stdout go to. Fiddling with sys.stdin/sys.stdout directly in the
# tests can be tricky.
STDIN = sys.stdin
STDOUT = sys.stdout


# Broker's basic types aren't JSON-serializable, so patch that up
# in this json.dumps() wrapper for JSON serialization of any object.
# Could go into utils.py, but it easier here to keep free of cyclic
# dependencies.
def json_dumps(obj: Any) -> str:
    def default(obj: bt.DataType) -> str:
        # Check specific Python types:
        if isinstance(obj, ipaddress.IPv4Address):
            return str(obj)
        if isinstance(obj, ipaddress.IPv6Address):
            return str(obj)

        # Specific zeek-client types (types.py):
        if isinstance(obj, Enum):
            return obj.to_json_data()

        # Specific brokertypes:
        if isinstance(obj, bt.Port):
            return str(obj.number)
        if isinstance(obj, bt.Timespan):
            return f"{obj._td.microseconds}us"
        # Fallback: assume the type's own Python representation is right.
        # json.dumps() will complain when that does not work.
        if isinstance(obj, bt.Type):
            return obj.to_py()  # type: ignore

        raise TypeError(f"cannot serialize {type(obj)} ({str(obj)})")

    indent = 2 if CONFIG.getboolean("client", "pretty_json") else None
    return json.dumps(obj, default=default, sort_keys=True, indent=indent)


def create_controller() -> Controller | None:
    try:
        ctl = Controller()
    except ControllerError as err:
        LOG.error(str(err))
        return None

    if not ctl.connect():
        return None

    return ctl


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="A Zeek management client",
        epilog="environment variables:\n\n"
        "    ZEEK_CLIENT_CONFIG_FILE:      "
        "Same as `--configfile` argument, but lower precedence.\n"
        "    ZEEK_CLIENT_CONFIG_SETTINGS:  "
        "Same as a space-separated series of `--set` arguments, but lower precedence.\n",
    )

    controller = (
        f"{CONFIG.get('controller', 'host')}:{CONFIG.get('controller', 'port')}"
    )

    parser.add_argument(
        "-c",
        "--configfile",
        metavar="FILE",
        default=CONFIG_FILE,
        help=f"Path to zeek-client config file. (Default: {CONFIG_FILE})",
    )
    parser.add_argument(
        "--controller",
        metavar="HOST:PORT",
        help=(
            f"Address and port of the controller, either of which may "
            f"be omitted (default: {controller})"
        ),
    )
    arg = parser.add_argument(
        "--set",
        metavar="SECTION.KEY=VAL",
        action="append",
        default=[],
        help="Adjust a configuration setting. Can use repeatedly. See show-settings.",
    )

    # This is for argcomplete users and has no effect otherwise.
    setattr(arg, "completer", CONFIG.completer)  # noqa: B010

    verbosity_group = parser.add_mutually_exclusive_group()
    verbosity_group.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Suppress informational output to stderr.",
    )
    verbosity_group.add_argument(
        "--verbose",
        "-v",
        action="count",
        help="Increase informational output to stderr. "
        "Repeat for more output (e.g. -vvv).",
    )

    parser.add_argument(
        "--version",
        action="store_true",
        help="Show version number and exit.",
    )

    command_parser = parser.add_subparsers(
        title="commands",
        dest="command",
        help="See `%(prog)s <command> -h` for per-command usage info.",
    )

    sub_parser = command_parser.add_parser(
        "deploy",
        help="Deploy a staged cluster configuration.",
    )
    sub_parser.set_defaults(run_cmd=cmd_deploy)

    sub_parser = command_parser.add_parser(
        "deploy-config",
        help="Upload a cluster configuration and deploy it.",
    )
    sub_parser.set_defaults(run_cmd=cmd_deploy_config)
    sub_parser.add_argument(
        "config",
        metavar="FILE",
        help='Cluster configuration file, "-" for stdin',
    )

    sub_parser = command_parser.add_parser(
        "get-config",
        help="Retrieve staged or deployed cluster configuration.",
    )
    sub_parser.set_defaults(run_cmd=cmd_get_config)
    sub_parser.add_argument(
        "--filename",
        "-f",
        metavar="FILE",
        default="-",
        help="Output file for the configuration, default stdout",
    )
    sub_parser.add_argument(
        "--as-json",
        action="store_true",
        help="Report in JSON instead of INI-style config file",
    )
    get_config_group = sub_parser.add_mutually_exclusive_group()
    get_config_group.add_argument(
        "--deployed",
        action="store_true",
        dest="deployed",
        default=False,
        help="Return deployed configuration",
    )
    get_config_group.add_argument(
        "--staged",
        action="store_false",
        dest="deployed",
        help="Return staged configuration (default)",
    )

    sub_parser = command_parser.add_parser(
        "get-id-value",
        help="Show the value of a given identifier in Zeek cluster nodes.",
    )
    sub_parser.set_defaults(run_cmd=cmd_get_id_value)
    sub_parser.add_argument(
        "id",
        metavar="IDENTIFIER",
        help="Name of the Zeek script identifier to retrieve.",
    )
    sub_parser.add_argument(
        "nodes",
        metavar="NODES",
        nargs="*",
        default=[],
        help="Name(s) of Zeek cluster nodes to query. When omitted, queries all nodes.",
    )

    sub_parser = command_parser.add_parser(
        "get-instances",
        help="Show instances connected to the controller.",
    )
    sub_parser.set_defaults(run_cmd=cmd_get_instances)

    sub_parser = command_parser.add_parser(
        "get-nodes",
        help="Show active Zeek nodes at each instance.",
    )
    sub_parser.set_defaults(run_cmd=cmd_get_nodes)

    sub_parser = command_parser.add_parser(
        "monitor",
        help="For troubleshooting: do nothing, just report events.",
    )
    sub_parser.set_defaults(run_cmd=cmd_monitor)

    sub_parser = command_parser.add_parser("restart", help="Restart cluster nodes.")
    sub_parser.set_defaults(run_cmd=cmd_restart)
    sub_parser.add_argument(
        "nodes",
        metavar="NODES",
        nargs="*",
        default=[],
        help="Name(s) of Zeek cluster nodes to restart. "
        "When omitted, restarts all nodes.",
    )

    sub_parser = command_parser.add_parser(
        "stage-config",
        help="Upload a cluster configuration for later deployment.",
    )
    sub_parser.set_defaults(run_cmd=cmd_stage_config)
    sub_parser.add_argument(
        "config",
        metavar="FILE",
        help='Cluster configuration file, "-" for stdin',
    )

    sub_parser = command_parser.add_parser(
        "show-settings",
        help="Show zeek-client's own configuration.",
    )
    sub_parser.set_defaults(run_cmd=cmd_show_settings)

    sub_parser = command_parser.add_parser(
        "test-timeout",
        help="Send timeout test event.",
    )
    sub_parser.set_defaults(run_cmd=cmd_test_timeout)
    sub_parser.add_argument(
        "--with-state",
        action="store_true",
        help="Make request stateful in the controller.",
    )

    return parser


def cmd_deploy(args: argparse.Namespace, controller: Controller | None = None) -> int:
    # The deploy-config command first stages a configuration and then calls this
    # function to deploy. We re-use its controller, passed to us.
    if controller is None:
        controller = create_controller()
    if controller is None:
        return 1

    resp, msg = controller.transact(DeployRequest, DeployResponse)

    if resp is None:
        LOG.error("no response received: %s", msg)
        return 1

    retval = 0
    json_data: dict[str, Any] = {
        "results": {},
        "errors": [],
    }

    for broker_data in resp.results:  # type: ignore
        res = Result.from_brokertype(broker_data)

        if not res.success:
            retval = 1

        if not res.success and res.node is None and res.error:
            # If a failure doesn't mention a node, it's either an agent
            # reporting an internal error, or the controller reporting a
            # config validation error.
            json_data["errors"].append(res.error)
            continue

        if res.success and res.node is None and res.instance is None and res.data:
            # It's success from the controller (since the instance field is
            # empty): the data field contains the ID of the deployed config.
            json_data["results"]["id"] = res.data
            continue

        # At this point we only expect responses from the agents:
        if res.instance is None:
            LOG.warning("skipping unexpected response %s", res)
            continue

        if res.node is None:
            # This happens when an agent handles deployment successfully, and
            # had no nodes to deploy. We skip this silently.
            continue

        # Everything else is node-specific results from agents.

        if "nodes" not in json_data["results"]:
            json_data["results"]["nodes"] = {}

        json_data["results"]["nodes"][res.node] = {
            "success": res.success,
            "instance": res.instance,
        }

        # If launching this node failed, we should have a NodeOutputs record as
        # data member in the result record. ("should", because on occasion
        # buffering in the node -> stem -> supervisor pipeline delays the
        # output.)
        if res.data:
            node_outputs = NodeOutputs.from_brokertype(res.data)
            json_data["results"]["nodes"][res.node]["stdout"] = node_outputs.stdout
            json_data["results"]["nodes"][res.node]["stderr"] = node_outputs.stderr

    print(json_dumps(json_data), file=STDOUT)
    return retval


def cmd_get_config(args: argparse.Namespace) -> int:
    controller = create_controller()
    if controller is None:
        return 1

    resp, msg = controller.transact(
        GetConfigurationRequest,
        GetConfigurationResponse,
        args.deployed,
    )

    if resp is None:
        LOG.error("no response received: %s", msg)
        return 1

    res = Result.from_brokertype(resp.result)

    if not res.success:
        msg = res.error if res.error else "no reason given"
        LOG.error(msg)
        return 1

    if not res.data:
        LOG.error("received result did not contain configuration data: %s", resp)
        return 1

    config = Configuration.from_brokertype(res.data)

    with (
        open(
            args.filename,
            "w",
            encoding="utf-8",
        )
        if args.filename and args.filename != "-"
        else STDOUT as hdl
    ):
        if args.as_json:
            hdl.write(json_dumps(config.to_json_data()) + "\n")
        else:
            cfp = config.to_config_parser()
            cfp.write(hdl)

    return 0


def cmd_get_id_value(args: argparse.Namespace) -> int:
    controller = create_controller()
    if controller is None:
        return 1

    resp, msg = controller.transact(
        GetIdValueRequest,
        GetIdValueResponse,
        args.id,
        set(args.nodes),
    )

    if resp is None:
        LOG.error("no response received: %s", msg)
        return 1

    json_data: dict[str, Any] = {
        "results": {},
        "errors": [],
    }

    # The Result records have both instance and node filled in, so use both for
    # ordering. While for the JSON serialization we can outsource the ordering
    # task to Python, for our error reporting it's up to us, and we want be
    # reproducible.

    results = [Result.from_brokertype(broker_data) for broker_data in resp.results]  # type: ignore

    for res in sorted(results):
        if not res.success:
            json_data["errors"].append(
                {
                    "source": res.node,
                    "error": res.error,
                },
            )
            continue

        # Upon success, we should always have res.node filled in. But guard anyway.
        if res.node:
            # res.data should be a string containing JSON rendered by Zeek's
            # to_json() BiF. Parse it into a data structure to render
            # seamlessly.
            if not isinstance(res.data, bt.String):
                json_data["errors"].append(
                    {
                        "source": res.node,
                        "error": f"invalid result data type {repr(res.data)}",
                    },
                )
                continue

            try:
                json_data["results"][res.node] = json.loads(res.data.to_py())
            except json.JSONDecodeError as err:
                json_data["errors"].append(
                    {
                        "source": res.node,
                        "error": f"JSON decode error: {err}",
                    },
                )
            continue

        json_data["errors"].append(
            {
                "error": f"result lacking node: {res.data}",
            },
        )

    print(json_dumps(json_data), file=STDOUT)
    return 0 if len(json_data["errors"]) == 0 else 1


def cmd_get_instances(_args: argparse.Namespace) -> int:
    controller = create_controller()
    if controller is None:
        return 1

    resp, msg = controller.transact(GetInstancesRequest, GetInstancesResponse)

    if resp is None:
        LOG.error("no response received: %s", msg)
        return 1

    res = Result.from_brokertype(resp.result)

    if not res.success:
        msg = res.error if res.error else "no reason given"
        LOG.error(msg)
        return 1

    if res.data is None:
        LOG.error("received result did not contain instance data: %s", resp)
        return 1

    json_data = {}

    # res.data is a (possibly empty) vector of Instances. Make the list of
    # instances easier to comprehend than raw Broker data: turn it into Instance
    # objects, then render these JSON-friendly.
    try:
        for inst in sorted([Instance.from_brokertype(inst) for inst in res.data]):  # type: ignore
            json_data[inst.name] = inst.to_json_data()
            json_data[inst.name].pop("name")
    except TypeError as err:
        LOG.error("instance data invalid: %s", err)

    print(json_dumps(json_data), file=STDOUT)
    return 0


def cmd_get_nodes(_args: argparse.Namespace) -> int:
    controller = create_controller()
    if controller is None:
        return 1

    resp, msg = controller.transact(GetNodesRequest, GetNodesResponse)

    if resp is None:
        LOG.error("no response received: %s", msg)
        return 1

    json_data: dict[str, Any] = {
        "results": {},
        "errors": [],
    }

    results = [Result.from_brokertype(broker_data) for broker_data in resp.results]  # type: ignore

    for res in sorted(results):
        if not res.success:
            json_data["errors"].append(
                {
                    "source": res.instance,
                    "error": res.error,
                },
            )
            continue

        if res.data is None:
            json_data["errors"].append(
                {
                    "source": res.instance,
                    "error": "result does not contain node status data",
                },
            )
            continue

        json_data["results"][res.instance] = {}

        # res.data is a NodeStatusVec
        try:
            nstats = [NodeStatus.from_brokertype(nstat_data) for nstat_data in res.data]  # type: ignore
            for nstat in sorted(nstats):
                # If either of the two role enums are "NONE", we make them
                # None. That way they stay in the reporting, but are more easily
                # distinguished from "actual" values.
                mgmt_role = (
                    nstat.mgmt_role if nstat.mgmt_role != ManagementRole.NONE else None
                )
                cluster_role = (
                    nstat.cluster_role
                    if nstat.cluster_role != ClusterRole.NONE
                    else None
                )

                json_data["results"][res.instance][nstat.node] = {
                    "state": nstat.state,
                    "mgmt_role": mgmt_role,
                    "cluster_role": cluster_role,
                }

                if nstat.pid is not None:
                    json_data["results"][res.instance][nstat.node]["pid"] = nstat.pid
                if nstat.port is not None:
                    json_data["results"][res.instance][nstat.node]["port"] = nstat.port
                if nstat.metrics_port is not None:
                    json_data["results"][res.instance][nstat.node]["metrics_port"] = (
                        nstat.metrics_port
                    )
        except TypeError as err:
            LOG.error("NodeStatus data invalid: %s", err)
            LOG.debug(traceback.format_exc())

    print(json_dumps(json_data), file=STDOUT)
    return 0 if len(json_data["errors"]) == 0 else 1


def cmd_monitor(_args: argparse.Namespace) -> int:
    controller = create_controller()
    if controller is None:
        return 1

    while True:
        resp, msg = controller.receive(timeout_secs=None)

        if resp is None:
            print(f"no response received: {msg}")
        else:
            print(f'received "{resp}"')


def cmd_restart(args: argparse.Namespace) -> int:
    controller = create_controller()
    if controller is None:
        return 1

    resp, msg = controller.transact(RestartRequest, RestartResponse, set(args.nodes))

    if resp is None:
        LOG.error("no response received: %s", msg)
        return 1

    json_data: dict[str, Any] = {
        "results": {},
        "errors": [],
    }

    # The Result records have both instance and node filled in, so use both for
    # ordering. While for the JSON serialization we can outsource the ordering
    # task to Python, for our error reporting it's up to us, and we want be
    # reproducible.

    results = [Result.from_brokertype(broker_data) for broker_data in resp.results]  # type: ignore

    for res in sorted(results):
        if not res.success and res.instance is None:
            # The controller generated this one, so add to errors section.
            json_data["errors"].append(
                {
                    "source": res.node,
                    "error": res.error,
                },
            )
            continue

        # Upon success, we should always have a node filled in. But guard anyway.
        if res.node:
            json_data["results"][res.node] = res.success
            continue

        json_data["errors"].append(
            {
                "error": f"result lacking node: {res}",
            },
        )

    print(json_dumps(json_data), file=STDOUT)
    return 0 if len(json_data["errors"]) == 0 else 1


def cmd_stage_config_impl(
    args: argparse.Namespace,
) -> tuple[int, dict[str, Any] | None, Controller | None]:
    """Internals of cmd_stage_config() to enable chaining with other commands.

    Returns a tuple of exit code, any JSON data to show to the user/caller, and
    the created controller object, if any.
    """
    if not args.config or (args.config != "-" and not os.path.isfile(args.config)):
        LOG.error("please provide a cluster configuration file.")
        return 1, None, None

    # We use a config parser to parse the cluster configuration. For instances,
    # we allow names without value to designate agents that connect to the
    # controller, like this:
    #
    # [instances]
    # foobar
    #
    # All other keys must have a value.
    config: Configuration | None = Configuration()
    cfp = configparser.ConfigParser(allow_no_value=True)

    if args.config == "-":
        cfp.read_file(STDIN)
    else:
        cfp.read(args.config)

    config = Configuration.from_config_parser(cfp)

    if config is None:
        LOG.error("configuration has errors, not sending")
        return 1, None, None

    controller = create_controller()
    if controller is None:
        return 1, None, None

    resp, msg = controller.transact(
        StageConfigurationRequest,
        StageConfigurationResponse,
        config.to_brokertype(),
    )

    if resp is None:
        LOG.error("no response received: %s", msg)
        return 1, None, controller

    retval = 0
    json_data: dict[str, Any] = {
        "results": {},
        "errors": [],
    }

    for broker_data in resp.results:  # type: ignore
        res = Result.from_brokertype(broker_data)

        if not res.success:
            retval = 1

            # Failures are config validation problems, trouble while
            # auto-assigning ports, or internal controller errors.
            # They should all come with error messages.
            json_data["errors"].append(res.error if res.error else "no reason given")
            continue

        if res.data:
            json_data["results"]["id"] = res.data

    return retval, json_data, controller


def cmd_stage_config(args: argparse.Namespace) -> int:
    ret, json_data, _ = cmd_stage_config_impl(args)

    if json_data:
        print(json_dumps(json_data), file=STDOUT)

    return ret


def cmd_deploy_config(args: argparse.Namespace) -> int:
    ret, json_data, controller = cmd_stage_config_impl(args)

    if ret != 0:
        if json_data:
            print(json_dumps(json_data), file=STDOUT)
        return ret

    return cmd_deploy(args, controller=controller)


def cmd_show_settings(_args: argparse.Namespace) -> int:
    CONFIG.write(STDOUT)
    return 0


def cmd_test_timeout(args: argparse.Namespace) -> int:
    controller = create_controller()
    if controller is None:
        return 1

    resp, msg = controller.transact(
        TestTimeoutRequest,
        TestTimeoutResponse,
        args.with_state,
    )

    if resp is None:
        LOG.error("no response received: %s", msg)
        return 1

    res = Result.from_brokertype(resp.result)
    print(json_dumps({"success": res.success, "error": res.error}), file=STDOUT)
    return 0



# Configuration options.
#


class Option:
    # Options category.
    USER = 1  # Standard user-configurable option.
    INTERNAL = 2  # internal, don't expose to user.
    AUTOMATIC = 3  # Set automatically, unlikely to be changed.

    def __init__(
        self, name, default, type, category, dontinit, description, legacy_name=None
    ):
        self.name = name
        self.default = default
        self.type = type
        self.category = category
        self.dontinit = dontinit
        self.description = description
        self.legacy_name = legacy_name

        if type == "string":
            if not isinstance(default, str):
                raise ValueError(f"option '{name}' default value must be string")
        else:
            if not isinstance(default, int):
                raise ValueError(f"option '{name}' default value must be int")


options = [
    # User options.
    Option(
        "Debug",
        0,
        "bool",
        Option.USER,
        False,
        "Enable extensive debugging output in spool/debug.log.",
    ),
    Option(
        "HaveNFS",
        0,
        "bool",
        Option.USER,
        False,
        "True if shared files are mounted across all nodes via NFS (see the FAQ_).",
    ),
    Option(
        "ClusterBackend",
        "ZeroMQ",
        "string",
        Option.USER,
        False,
        "Supported values: Broker or ZeroMQ",
    ),
    Option(
        "SaveTraces",
        0,
        "bool",
        Option.USER,
        False,
        "True to let backends capture short-term traces via '-w'. These are not archived but might be helpful for debugging.",
    ),
    Option(
        "StopTimeout",
        60,
        "int",
        Option.USER,
        False,
        "The number of seconds to wait before sending a SIGKILL to a node which was previously issued the 'stop' command but did not terminate gracefully.",
    ),
    Option(
        "CommTimeout",
        10,
        "int",
        Option.USER,
        False,
        "The number of seconds to wait before assuming Broker communication events have timed out.",
    ),
    Option(
        "ControlTopic",
        "zeek/control",
        "string",
        Option.USER,
        False,
        "The Broker topic name used for sending and receiving control messages to Zeek processes.",
    ),
    Option(
        "CommandTimeout",
        60,
        "int",
        Option.USER,
        False,
        "The number of seconds to wait for a command to return results.",
    ),
    Option(
        "ZeekPort",
        27760,
        "int",
        Option.USER,
        False,
        "The TCP port number that Zeek will listen on. For a cluster configuration, each node in the cluster will automatically be assigned a subsequent port to listen on.",
        "BroPort",
    ),
    Option(
        "MetricsAddress",
        "127.0.0.1",
        "string",
        Option.USER,
        False,
        "The address on which the Prometheus HTTP server listens on (Telemetry::metrics_address).",
    ),
    Option(
        "MetricsPort",
        9991,
        "int",
        Option.USER,
        False,
        "The TCP port number that Zeek will listen on for Prometheus telemetry. For a cluster configuration, each node in the cluster will automatically be assigned a subsequent port to listen on. Setting this to 0 will disable telemetry on all nodes.",
    ),
    Option(
        "LogRotationInterval",
        3600,
        "int",
        Option.USER,
        False,
        "The frequency of log rotation in seconds for the manager/standalone node (zero to disable rotation). This overrides the Zeek script variable Log::default_rotation_interval.",
    ),
    Option(
        "LogDir",
        "${ZeekBase}/logs",
        "string",
        Option.USER,
        False,
        "Directory for archived log files.",
    ),
    Option(
        "MakeArchiveName",
        "${ZeekBase}/share/zeekctl/scripts/make-archive-name",
        "string",
        Option.USER,
        False,
        "Script to generate filenames for archived log files.",
    ),
    Option(
        "CompressLogs",
        1,
        "bool",
        Option.USER,
        False,
        "True to compress archived log files.",
    ),
    Option(
        "CompressLogsInFlight",
        0,
        "int",
        Option.USER,
        False,
        "Set to greater than zero to compress archived log files as they're created instead of during rotation.  The value indicates the compression level to use between 1 and 9 (values of 6 or 7 are a typical choice to bias slightly more towards better compression at cost of performance). If this is enabled, the CompressLogs, and CompressCmd arguments will be ignored as the files are compressed automatically by Zeek.",
    ),
    Option(
        "CompressCmd",
        "gzip",
        "string",
        Option.USER,
        False,
        "If archived logs will be compressed, the command to use for that. The specified command must compress its standard input to standard output.",
    ),
    Option(
        "CompressExtension",
        "gz",
        "string",
        Option.USER,
        False,
        "If archived logs will be compressed, the file extension to use on compressed log files. When specifying a file extension, don't include the period character (e.g., specify 'gz' instead of '.gz').",
    ),
    Option(
        "PrivateAddressSpaceIsLocal",
        1,
        "bool",
        Option.USER,
        False,
        "This flag, enabled by default, controls whether Zeek should automatically consider private address space as local to your site. This is the zeekctl equivalent of Zeek's 'Site::private_address_space_is_local' setting. Setting this to 0 separates local and private address spaces, and you need to list any private address space explicitly in your 'network.cfg' for it to be considered local.",
    ),
    Option(
        "SendMail",
        "/usr/sbin/sendmail",
        "string",
        Option.USER,
        False,
        "Location of the sendmail binary.  Make this string blank to prevent email from being sent. The default value is configuration-dependent and determined automatically by CMake at configure-time. This overrides the Zeek script variable Notice::sendmail.",
    ),
    Option(
        "MailSubjectPrefix",
        "[Zeek]",
        "string",
        Option.USER,
        False,
        "General Subject prefix for mails. This overrides the Zeek script variable Notice::mail_subject_prefix.",
    ),
    Option(
        "MailReplyTo",
        "",
        "string",
        Option.USER,
        False,
        "Reply-to address for zeekctl-generated mails.",
    ),
    Option(
        "MailTo",
        "<user>",
        "string",
        Option.USER,
        True,
        "Destination address for non-alarm mails. This overrides the Zeek script variable Notice::mail_dest.",
    ),
    Option(
        "MailFrom",
        "Zeek <zeek@localhost>",
        "string",
        Option.USER,
        True,
        "Originator address for mails. This overrides the Zeek script variable Notice::mail_from.",
    ),
    Option(
        "MailAlarmsTo",
        "${MailTo}",
        "string",
        Option.USER,
        True,
        "Destination address for alarm summary mails. Default is to use the same address as MailTo. This overrides the Zeek script variable Notice::mail_dest_pretty_printed.",
    ),
    Option(
        "MailAlarmsInterval",
        86400,
        "int",
        Option.USER,
        False,
        "The frequency (in seconds) of sending alarm summary mails (zero to disable). This overrides the Zeek script variable Log::default_mail_alarms_interval.",
    ),
    Option(
        "MailConnectionSummary",
        1,
        "bool",
        Option.USER,
        False,
        "True to mail connection summary reports each log rotation interval (if false, then connection summary reports will still be generated and archived, but they will not be mailed). However, this option has no effect if the trace-summary script is not available.",
    ),
    Option(
        "MailHostUpDown",
        1,
        "bool",
        Option.USER,
        False,
        "True to enable sending mail when zeekctl cron notices the availability of a host in the cluster to have changed.",
    ),
    Option(
        "MailArchiveLogFail",
        1,
        "bool",
        Option.USER,
        False,
        "True to enable sending mail when log files fail to be archived.",
    ),
    Option(
        "MailReceivingPackets",
        1,
        "bool",
        Option.USER,
        False,
        "True to enable sending mail when zeekctl cron notices that an interface is not receiving any packets (note that such mail is not sent when StatsLogEnable is 0).",
    ),
    Option(
        "MinDiskSpace",
        5,
        "int",
        Option.USER,
        False,
        "Minimum percentage of disk space available before zeekctl cron mails a warning.  If this value is 0, then no warning will be sent.",
    ),
    Option(
        "StatsLogEnable",
        1,
        "bool",
        Option.USER,
        False,
        "True to enable ZeekControl to write statistics to the stats.log file.",
    ),
    Option(
        "StatsLogExpireInterval",
        0,
        "int",
        Option.USER,
        False,
        "Number of days entries in the stats.log file are kept (zero means never expire).",
    ),
    Option(
        "CrashExpireInterval",
        0,
        "int",
        Option.USER,
        False,
        "Number of days that crash directories are kept (zero means never expire).",
    ),
    Option(
        "LogExpireInterval",
        "0",
        "string",
        Option.USER,
        False,
        "Time interval that archived log files are kept (a value of 0 means log files never expire).  The time interval is expressed as an integer followed by one of the following time units: day, hr, min.",
    ),
    Option(
        "KeepLogs",
        "",
        "string",
        Option.USER,
        False,
        "A space-separated list of filename shell patterns of expired log files to keep (empty string means don't keep any expired log files). The filename shell patterns are not regular expressions and do not include any directories. For example, specifying 'conn.* dns*' will prevent any expired log files with filenames starting with 'conn.' or 'dns' from being removed. Finally, note that this option is ignored if log files never expire.",
    ),
    Option(
        "ZeekArgs",
        "",
        "string",
        Option.USER,
        False,
        'Additional arguments to pass to Zeek on the command-line (e.g. zeekargs=-f "tcp port 80").',
        "BroArgs",
    ),
    Option(
        "MemLimit",
        "unlimited",
        "string",
        Option.USER,
        False,
        "Maximum amount of memory for Zeek processes to use (in KB, or the string 'unlimited').",
    ),
    Option(
        "Env_Vars",
        "",
        "string",
        Option.USER,
        False,
        "A comma-separated list of environment variables (e.g. env_vars=VAR1=123, VAR2=456) to set on all nodes immediately before starting Zeek.  Node-specific values (specified in the node configuration file) override these global values.",
    ),
    Option(
        "TimeFmt",
        "%d %b %H:%M:%S",
        "string",
        Option.USER,
        False,
        "Format string to print date/time specifications (see 'man strftime').",
    ),
    Option(
        "Prefixes",
        "local",
        "string",
        Option.USER,
        False,
        "Additional script prefixes for Zeek, separated by colons. Use this instead of @prefix.",
    ),
    Option(
        "SitePolicyScripts",
        "local.zeek",
        "string",
        Option.USER,
        False,
        "Space-separated list of local policy files that will be automatically loaded for all Zeek instances.  Scripts listed here do not need to be explicitly loaded from any other policy scripts.",
    ),
    Option(
        "StatusCmdShowAll",
        0,
        "bool",
        Option.USER,
        False,
        "True to have the status command show all output, or False to show only some of the output (peer information will not be collected or shown, so the command will run faster).",
    ),
    Option(
        "StopWait",
        0,
        "bool",
        Option.USER,
        False,
        "True to force the stop command to wait for the post-terminate script to finish, or False to let post-terminate finish in the background.",
    ),
    Option(
        "CronCmd",
        "",
        "string",
        Option.USER,
        False,
        "A custom command to run everytime the cron command has finished.",
    ),
    Option(
        "PFRINGClusterID",
        21,
        "int",
        Option.USER,
        False,
        "If PF_RING flow-based load balancing is desired, this is where the PF_RING cluster id is defined.  In order to use PF_RING, the value of this option must be non-zero.",
    ),
    Option(
        "PFRINGClusterType",
        "4-tuple",
        "string",
        Option.USER,
        False,
        "If PF_RING flow-based load balancing is desired, this is where the PF_RING cluster type is defined.  Allowed values are: 2-tuple, 4-tuple, 5-tuple, tcp-5-tuple, 6-tuple, inner-2-tuple, inner-4-tuple, inner-5-tuple, inner-tcp-5-tuple, or inner-6-tuple.  Zeek must be linked with PF_RING's libpcap wrapper and PFRINGClusterID must be non-zero for this option to work.",
    ),
    Option(
        "PFRINGFirstAppInstance",
        0,
        "int",
        Option.USER,
        False,
        "The first application instance for a PF_RING dnacluster interface to use.  Zeekctl will start at this application instance number and increment for each new process running on that DNA cluster.  Zeek must be linked with PF_RING's libpcap wrapper, PFRINGClusterID must be non-zero, and you must be using PF_RING+DNA and libzero for this option to work.",
    ),
    Option(
        "PcapSnaplen",
        9216,
        "int",
        Option.AUTOMATIC,
        False,
        "Number of bytes per packet to capture from live interfaces via libpcap.",
    ),
    Option(
        "PcapBufsize",
        128,
        "int",
        Option.AUTOMATIC,
        False,
        "Number of Mbytes to provide as buffer space when capturing from live interfaces via libpcap.",
    ),
    Option(
        "TimeMachineHost",
        "",
        "string",
        Option.USER,
        False,
        "If the manager should connect to a Time Machine, the address of the host it is running on.",
    ),
    Option(
        "TimeMachinePort",
        "47757/tcp",
        "string",
        Option.USER,
        False,
        "If the manager should connect to a Time Machine, the port it is running on (in Zeek syntax, e.g., 47757/tcp).",
    ),
    Option(
        "UseWebSocket",
        1,
        "bool",
        Option.USER,
        False,
        "True to switch Zeekctl to use the /v1/messages/json WebSocket API for communicating with the cluster. Setting this to true will spawn a WebSocket server on the manager node. This requires the Python websockets package version 11.0 or higher.",
    ),
    Option(
        "WebSocketHost",
        "127.0.0.1",
        "string",
        Option.USER,
        False,
        "If UseWebSocket is true, the WebSocket address the manager will listen on.",
    ),
    Option(
        "WebSocketPort",
        27759,
        "int",
        Option.USER,
        False,
        "If UseWebSocket is true, the WebSocket port the manager will listen on.",
    ),
    Option(
        "WebSocketUrl",
        "",
        "string",
        Option.USER,
        False,
        "Override the WebSocket URL used by zeekctl. Only used when UseWebSocket is true.",
    ),
    Option(
        "WebSocketTimeout",
        10,
        "int",
        Option.USER,
        False,
        "The timeout for WebSocket operations (connect, close and waiting for events).",
    ),
    # Automatically set.
    Option(
        "ZeekBase",
        "",
        "string",
        Option.AUTOMATIC,
        True,
        "Base path of zeekctl installation on all nodes.",
        "BroBase",
    ),
    Option("Version", "", "string", Option.AUTOMATIC, True, "Version of the zeekctl."),
    Option(
        "StandAlone",
        0,
        "bool",
        Option.AUTOMATIC,
        True,
        "True if running in stand-alone mode (see elsewhere).",
    ),
    Option(
        "OS",
        "",
        "string",
        Option.AUTOMATIC,
        True,
        "Name of operating system as reported by uname.",
    ),
    Option("Time", "", "string", Option.AUTOMATIC, True, "Path to time binary."),
    Option(
        "LogExpireMinutes",
        0,
        "int",
        Option.AUTOMATIC,
        True,
        "Time interval (in minutes) that archived log files are kept (0 means they never expire).  Users should never modify this value (see the LogExpireInterval option).",
    ),
    Option(
        "BinDir",
        "${ZeekBase}/bin",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for executable files.",
    ),
    Option(
        "Zeek",
        "${BinDir}/zeek",
        "string",
        Option.AUTOMATIC,
        False,
        "Path to Zeek binary.",
        "Bro",
    ),
    Option(
        "ScriptsDir",
        "${ZeekBase}/share/zeekctl/scripts",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for executable scripts shipping as part of zeekctl.",
    ),
    Option(
        "PostProcDir",
        "${ZeekBase}/share/zeekctl/scripts/postprocessors",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for log postprocessors.",
    ),
    Option(
        "HelperDir",
        "${ZeekBase}/share/zeekctl/scripts/helpers",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for zeekctl helper scripts.",
    ),
    Option(
        "CfgDir",
        "${ZeekBase}/etc",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for configuration files.",
    ),
    Option(
        "SpoolDir",
        "${ZeekBase}/spool",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for run-time data.",
    ),
    Option(
        "BrokerDBDir",
        "${ZeekBase}/spool/brokerstore",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for data stores of persistent Broker-backed tables.",
    ),
    Option(
        "FileExtractDir",
        "${SpoolDir}/extract_files",
        "string",
        Option.AUTOMATIC,
        False,
        "Default base directory for file extraction. The FileExtract module's prefix will be set to this value with Cluster::node appended.",
    ),
    Option(
        "PolicyDir",
        "${ZeekScriptDir}",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for standard policy files.",
    ),
    Option(
        "StaticDir",
        "${ZeekBase}/share/zeekctl",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for static, arch-independent files.",
    ),
    Option(
        "LibDir", "", "string", Option.AUTOMATIC, False, "Directory for library files."
    ),
    # XXX Do we still need the following? --cpk
    Option(
        "LibDir64",
        "${ZeekBase}/lib64",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for 64-bit architecture library files.",
    ),
    Option(
        "LibDirInternal",
        "",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for ZeekControl's Python module.",
    ),
    Option(
        "TmpDir",
        "${SpoolDir}/tmp",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory for temporary data.",
    ),
    Option(
        "TmpExecDir",
        "${SpoolDir}/tmp",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory where binaries are copied before execution.  This option is ignored if HaveNFS is 0.",
    ),
    Option(
        "StatsDir",
        "${LogDir}/stats",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory where statistics are kept.",
    ),
    Option(
        "PluginDir",
        "${LibDirInternal}/zeekctl/plugins",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory where standard zeekctl plugins are located.",
    ),
    Option(
        "PluginZeekDir",
        "${LibDir}/zeek/plugins",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory where Zeek plugins are located.  ZeekControl will search this directory tree for zeekctl plugins that are provided by any Zeek plugin.",
        "PluginBroDir",
    ),
    Option(
        "TraceSummary",
        "${bindir}/trace-summary",
        "string",
        Option.AUTOMATIC,
        False,
        "Path to trace-summary script (empty if not available). Make this string blank to disable the connection summary reports.",
    ),
    Option(
        "CapstatsPath",
        "${bindir}/capstats",
        "string",
        Option.AUTOMATIC,
        False,
        "Path to capstats binary; empty if not available.",
    ),
    Option(
        "NodeCfg",
        "${CfgDir}/node.cfg",
        "string",
        Option.AUTOMATIC,
        False,
        "Node configuration file.",
    ),
    Option(
        "LocalNetsCfg",
        "${CfgDir}/networks.cfg",
        "string",
        Option.AUTOMATIC,
        False,
        "File defining the local networks.",
    ),
    Option(
        "StateFile",
        "${SpoolDir}/state.db",
        "string",
        Option.AUTOMATIC,
        False,
        "File storing the current zeekctl state.",
    ),
    Option(
        "LockFile",
        "${SpoolDir}/lock",
        "string",
        Option.AUTOMATIC,
        False,
        "Lock file preventing concurrent shell operations.",
    ),
    Option(
        "DebugLog",
        "${SpoolDir}/debug.log",
        "string",
        Option.AUTOMATIC,
        False,
        "Log file for debugging information.",
    ),
    Option(
        "StatsLog",
        "${SpoolDir}/stats.log",
        "string",
        Option.AUTOMATIC,
        False,
        "Log file for statistics.",
    ),
    Option(
        "DefaultStoreDir",
        "${SpoolDir}/stores",
        "string",
        Option.AUTOMATIC,
        False,
        "Default directory where Broker data stores will be written if user has not provided further customizations on a per-store basis.",
    ),
    Option(
        "SitePolicyPath",
        "${PolicyDir}/site",
        "string",
        Option.USER,
        False,
        "Directories to search for local (i.e., site-specific) policy files, separated by colons. For each such directory, all files and subdirectories are copied to PolicyDirSiteInstall during zeekctl 'install' or 'deploy' (however, if the same file or subdirectory is found in more than one such directory, then only the first one encountered will be used).",
    ),
    Option(
        "SitePluginPath",
        "",
        "string",
        Option.USER,
        False,
        "Directories to search for custom plugins (i.e., plugins that are not included with zeekctl), separated by colons.",
    ),
    Option(
        "PolicyDirSiteInstall",
        "${SpoolDir}/installed-scripts-do-not-touch/site",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory where the shell copies local (i.e., site-specific) policy scripts when installing.",
    ),
    Option(
        "PolicyDirSiteInstallAuto",
        "${SpoolDir}/installed-scripts-do-not-touch/auto",
        "string",
        Option.AUTOMATIC,
        False,
        "Directory where the shell copies auto-generated local policy scripts when installing.",
    ),
    # Internal, not documented.
    Option(
        "ZeekCtlConfigDir",
        "${SpoolDir}",
        "string",
        Option.INTERNAL,
        False,
        """Directory where the shell copies the zeekctl-config.sh
           configuration file. If this is changed, the symlink created in
           CMakeLists.txt must be adapted as well.""",
        "BroCtlConfigDir",
    ),
    Option(
        "ClusterTopicSeparator",
        "",
        "string",
        Option.INTERNAL,
        False,
        "Separator used for joining node topics. This is set by cluster backend plugins.",
    ),
]


def print_options(category):
    out = ""
    err = ""

    for opt in sorted(options, key=lambda o: o.name):
        if opt.category != category:
            continue

        if not opt.type:
            err += f"no type given for {opt.name}\n"

        if opt.type == "string":
            if opt.default:
                opt.default = f'"{opt.default}"'
            else:
                opt.default = "_empty_"

        default = f", default {opt.default}"

        default = default.replace("{", "\\{")
        description = opt.description.replace("{", "\\{")

        out += f".. _{opt.name}:\n\n*{opt.name}* ({opt.type}{default})\n    {description}\n\n"

    return (out, err)



# This plugin introduces a new load balancing method "custom", which does
# nothing by default.  To support packet source plugins, it allows to set a
# prefix and suffix for the interface name.  For example, add the following
# line to zeekctl.cfg to enable the AF_PACKET plugin:
# lb_custom.InterfacePrefix = af_packet::

import ZeekControl.plugin


class LBCustom(ZeekControl.plugin.Plugin):
    def __init__(self):
        super().__init__(apiversion=1)

    def name(self):
        return "lb_custom"

    def pluginVersion(self):
        return 1

    def init(self):
        useplugin = False

        for nn in self.nodes():
            if nn.type != "worker" or nn.lb_method != "custom":
                continue

            useplugin = True

            prefix = self.getOption("InterfacePrefix")
            suffix = self.getOption("InterfaceSuffix")
            nn.interface = f"{prefix}{nn.interface}{suffix}"

        return useplugin

    def options(self):
        custom_options = [
            (
                "InterfacePrefix",
                "string",
                "",
                "Prefix to prepend to the configured interface name.",
            ),
            (
                "InterfaceSuffix",
                "string",
                "",
                "Suffix to append to the configured interface name.",
            ),
        ]
        return custom_options



"""
A plugin warning about the ZeekPort change coming with Zeek 5.2.
"""

import textwrap

import ZeekControl.cmdresult
import ZeekControl.plugin


class ZeekPortWarningPlugin(ZeekControl.plugin.Plugin):
    def __init__(self):
        super().__init__(apiversion=1)

    def name(self):
        return "ZeekPort warning"

    def prefix(self):
        return "zeek_port_warning"

    def pluginVersion(self):
        return 1

    def options(self):
        return [("disable", "bool", False, "Disable the ZeekPort warning.")]

    def warning_banner(self, text):
        for i, line in enumerate(text.splitlines()):
            if i == 0 and not line:
                continue
            self.message("WARNING: " + line)

    def init(self):
        """
        Print a warning about the change of the ZeekPort option.

        We're just assuming that if it's 47760, the default was selected.
        """
        if self.getOption("disable"):
            return

        zeek_port = self.getGlobalOption("ZeekPort")
        uses_old_default_port = zeek_port == 47760

        if uses_old_default_port:
            self.warning_banner("*" * 80)
            if self.getGlobalOption("os") == "Linux":
                self.warning_banner(
                    textwrap.dedent(
                        """
                    You're using Linux with the default ZeekPort setting 47760. This configuration
                    is know to cause persistent worker failures with error messages as follows:

                        error in <...>/cluster/setup-connections.zeek, lines 94-96: Failed to listen on INADDR_ANY:47764 (...)

                    """
                    )
                )

            self.warning_banner(
                textwrap.dedent("""
                Starting with Zeek 5.2, the default ZeekPort used by zeekctl will
                change from 47760 to 27760 in order to avoid potential port collisions
                with other processes due to 47760 falling right into Linux's default
                ephemeral port range.

                Consider changing the ZeekPort option in your zeekctl.cfg to 27760
                now to prepare for this change. Doing so will silence this warning.

                    ZeekPort = 27760

                Note, if you're employing strict firewall rules between Zeek nodes,
                you'll likely need to update these rules. If you're using Zeek on
                a single physical host, no further action should be required.
                If possible do test the change in a non-production environment.

                To silence this warning without changing the ZeekPort option,
                set zeek_port_warning.disable = 1 in zeekctl.cfg.

                See the following PR for more details:
                    https://github.com/zeek/zeekctl/pull/41

                Feel free to reach out on zeekorg.slack.com or community.zeek.org if
                you have any questions around this change.
            """)
            )

            self.warning_banner("*" * 80)



"""This module provides specializes the Python ConfigParser class, pre-defining
a set of config settings for zeek-client and includes ways to override these
from the environment and the command line.
"""

import argparse
import configparser
import os
import shlex
from typing import Any

from .consts import CONFIG_FILE
from .logs import LOG


class Config(configparser.ConfigParser):
    """zeek-client configuration settings.

    A specialized ConfigParser that hardwires defaults for select values.
    Three levels of overrides apply, if provided:

    (1) first, the config file, if available

    (2) the ZEEK_CLIENT_CONFIG_SETTINGS environment variable may contain a
        series of <section.key>=<val> assignments

    (3) Any --set <section.key>=<val> arguments apply final overrides
    """

    def __init__(self) -> None:
        super().__init__()
        self.reset()

    def reset(self) -> None:
        self.read_dict(
            {
                "client": {
                    # The default timeout for request state is 15 seconds on the
                    # Zeek side, so by making it larger here we ensure that timeout
                    # events can fire & propagate in Zeek before we give up here.
                    "request_timeout_secs": 20,
                    # Successful peering requires a successful WebSocket connection
                    # to the controller and the successful exchange of peering
                    # handshake and response. We retry both, counting connection as
                    # well as handshake attempts toward this total:
                    "peering_attempts": 10,
                    # How long the client's Broker's endpoint should wait internally
                    # until it retries a peering upon connection or when the
                    # connection goes away. Its default is 10 seconds; we dial that
                    # down to be more interactive.
                    "peering_retry_delay_secs": 1.0,
                    # The way zeek-client reports informational messages on stderr
                    "rich_logging_format": False,
                    # Whether we pretty-print JSON output by default.
                    "pretty_json": True,
                    # Default output verbosity level:
                    #   0   permits errors
                    #   1   also warnings
                    #   2   also informational messages
                    #   3   also debug messages
                    #   4   also trace websocket I/O
                    #   5+  no additional effect
                    "verbosity": 0,
                },
                "controller": {
                    # Default host name/address where we contact the controller.
                    "host": "127.0.0.1",
                    # Default WebSocket port of the controller.
                    "port": 2149,
                },
                "ssl": {
                    # Security settings of the connection to the controller.
                    # Connections are unencrypted by default. To enable Broker's
                    # default mode of encryption with unverified endpoints, just
                    # set "enable" to true. Providing certificate, private
                    # key, and possibly CA & passphrase secure the connection
                    # properly.
                    #
                    # Boolean to indicate whether to use SSL at all. By default
                    # the client uses plaintext communication.
                    "enable": False,
                    # Path to a file containing a X.509 certificate in PEM format.
                    "certificate": "",
                    # Path to a file containing the private key for the certificate,
                    # in PEM format.
                    "keyfile": "",
                    # Path to a file containing concatenated, trusted certificates,
                    # in PEM format.
                    "cafile": "",
                    # Path to an OpenSSL-style directory of trusted certificates.
                    "capath": "",
                    # A passphrase to decrypt the private key, if required.
                    "passphrase": "",
                },
            },
        )

    def update_from_file(self, config_file: str = CONFIG_FILE) -> None:
        self.read(config_file)

    def update_from_env(self) -> None:
        for item in shlex.split(os.getenv("ZEEK_CLIENT_CONFIG_SETTINGS") or ""):
            try:
                self.apply(item)
            except ValueError:
                LOG.error(
                    'config item "%s" in ZEEK_CLIENT_CONFIG_SETTINGS '
                    "invalid. Please use <section.key>=<val>.",
                    item,
                )

    def update_from_args(self, args: argparse.Namespace) -> None:
        for item in args.set:
            try:
                self.apply(item)
            except ValueError:
                LOG.error(
                    'config item "%s" invalid. Please use <section.key>=<val>.',
                    item,
                )

        # The `--controller` argument is a shortcut for two `--set` arguments that
        # set controller host and port, so update these manually:
        if args.controller:
            if ":" not in args.controller:
                host = args.controller
                port = ""
            else:
                (host, _, port) = args.controller.rpartition(":")
                if host.count(":") >= 1:
                    # We likely have an IPv6 address
                    if not host.startswith("["):
                        raise ValueError(
                            "IPv6 addresses must be surrounded by brackets: [<ipv6>]:<port>"
                        )
                    if port.endswith("]"):
                        host = host + ":" + port
                        port = ""

            if port == "":
                # It's just a hostname
                self.set("controller", "host", host)
            elif host == "":
                # It's just a port (as ":<port>")
                self.set("controller", "port", port)
            else:
                self.set("controller", "host", host)
                self.set("controller", "port", port)

        # --verbose/-v/-vvv etc set a numeric verbosity level:
        if args.verbose:
            self.set("client", "verbosity", str(args.verbose))

    def apply(self, item: str) -> None:
        """This is equivalent to set(), but works via a single <section.key>=<val> string."""
        try:
            identifier, val = item.split("=", 1)
            section, key = identifier.split(".", 1)
            if not self.has_section(section):
                self.add_section(section)
            self.set(section, key, val)
        except ValueError as err:
            raise ValueError(f'config item "{item}" invalid') from err

    def completer(self, **_kwargs: Any) -> list[str]:
        """A completer suitable for argcomplete."""
        ret = []

        for section in self.sections():
            for key, val in self.items(section):
                ret.append(section + "." + key + "=" + val)

        return sorted(ret)


CONFIG = Config()



"""This package defines a Python interface for interacting with a Zeek cluster
controller. It provides a Controller class for sending and retrieving events, a
Zeek event abstraction on top of Broker, and various types that mirror those
defined in Zeek's policy/management/frameworks/management/types.zeek, needed
for request/response events.
"""

from . import (
    brokertypes,
    cli,
    config,
    consts,
    controller,
    events,
    logs,
    ssl,
    types,
    utils,
)
from .config import CONFIG
from .consts import (
    CONFIG_FILE,
    CONTROLLER_TOPIC,
)
from .logs import LOG

__version__ = "1.4.0-12"
__all__ = [
    "brokertypes",
    "cli",
    "config",
    "consts",
    "controller",
    "events",
    "logs",
    "ssl",
    "types",
    "utils",
    "CONFIG",
    "CONTROLLER_TOPIC",
    "CONFIG_FILE",
    "LOG",
]



# These modules provides a set of functions to execute actions on a host.
# If the host is local, it's done direcly; if it's remote we log in via SSH.

import logging
import os
import shutil
import subprocess

from ZeekControl import ssh_runner, util


# Copy src to dstdir, preserving permission bits and file type.  The src
# file type can be symlink, regular file, or directory (directories are copied
# recursively).  If the target pathname already exists, it is not clobbered.
def install(src, dstdir, cmdout):
    if not os.path.lexists(src):
        cmdout.error(f"pathname not found: {src}")
        return False

    dst = os.path.join(dstdir, os.path.basename(src))
    if os.path.lexists(dst):
        # Do not clobber existing files/dirs (this is not an error)
        return True

    logging.debug("cp %s %s", src, dstdir)

    try:
        if os.path.islink(src):
            target = os.readlink(src)
            os.symlink(target, dst)
        elif os.path.isfile(src):
            shutil.copy2(src, dstdir)
        elif os.path.isdir(src):
            shutil.copytree(src, dst, symlinks=True)
        else:
            cmdout.error(f"failed to copy {src}: not a file, dir, or symlink")
            return False
    except OSError:
        # Python 2.6 has a bug where this may fail on NFS. So we just
        # ignore errors.
        pass
    except OSError as err:
        cmdout.error(f"failed to copy: {err}")
        return False

    return True


# rsyncs paths from localhost to destination hosts.
def sync(nodes, paths, cmdout):
    result = True
    cmds = []
    for n in nodes:
        args = [
            "-rRl",
            "--delete",
            '--rsh="ssh -o BatchMode=yes -o LogLevel=error -o ConnectTimeout=30"',
        ]
        dst = [f"{util.format_rsync_addr(n.addr)}:/"]
        args += paths + dst
        cmdline = "rsync {}".format(" ".join(args))
        cmds += [(n, cmdline, "", None)]

    for id, success, output in run_localcmds(cmds):
        if not success:
            cmdout.error(f"rsync to {id.addr} failed: {output}")
            result = False

    return result


# Runs command locally and returns tuple (success, output)
# with success being true if the command terminated with exit code 0,
# and output is a string containing the combined stdout/stderr output of the
# command.
# The "env" is a space-separated string of environment variables to set,
# and "inputtext" is a string to send to stdin.
def run_localcmd(cmd, env=None, inputtext=None):
    proc = _run_localcmd_init("single", cmd, env)
    return _run_localcmd_wait(proc, inputtext)


# Same as run_localcmd() but runs a set of local commands in parallel.
# Cmds is a list of (id, cmd, envs, inputtext) tuples, where id is
# an arbitrary cookie identifying each command.
# Returns a list of (id, success, output) tuples.
def run_localcmds(cmds):
    results = []
    running = []

    for id, cmd, envs, inputtext in cmds:
        proc = _run_localcmd_init(id, cmd, envs)
        running += [(id, proc, inputtext)]

    for id, proc, inputtext in running:
        success, output = _run_localcmd_wait(proc, inputtext)
        results += [(id, success, output)]

    return results


def _run_localcmd_init(id, cmd, env):
    if env:
        cmdline = env + " " + cmd
    else:
        cmdline = cmd

    logging.debug(cmdline)

    # os.setsid makes sure that the child process doesn't receive our CTRL-Cs.
    proc = subprocess.Popen(
        [cmdline],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        close_fds=True,
        shell=True,
        preexec_fn=os.setsid,
    )

    return proc


def _run_localcmd_wait(proc, inputtext):
    if inputtext:
        inputtext = inputtext.encode()

    # Note: "output" is combined stdout/stderr output.
    output, _ = proc.communicate(inputtext)
    rc = proc.returncode

    output = output.decode()

    logging.debug("exit status: %d", rc)

    return (rc == 0, output)


# FIXME: This is an ugly hack. The __del__ method produces
# strange unhandled exceptions in the child at termination
# of the main process. Not sure if disabling the cleanup
# altogether is a good thing but right now that's the
# only fix I can come up with.
def _emptyDel(self):
    pass


subprocess.Popen.__del__ = _emptyDel


class Executor:
    def __init__(self, config):
        self.config = config
        self.sshrunner = ssh_runner.MultiMasterManager(config.localaddrs)

    def finish(self):
        self.sshrunner.shutdown_all()

    # Run commands in parallel on one or more hosts.
    #
    # cmds:  a list of the form: [ (node, cmd, args), ... ]
    #   where "cmd" is a string, "args" is a list of strings.
    # shell:  if True, then the "cmd" (and "args") will be interpreted by a
    #   shell.
    # helper:  if True, then the "cmd" will be modified to specify the full
    #   path to the zeekctl helper script.
    #
    # Returns a list of results: [(node, success, output), ...]
    #   where "success" is a boolean (True if command's exit status was zero),
    #   and "output" is a string containing the command's stdout followed by
    #   stderr, or an error message if no result was received (this could occur
    #   upon failure to communicate with remote host, or if the command being
    #   executed did not finish before the timeout).
    def run_cmds(self, cmds, shell=False, helper=False):
        results = []

        if not cmds:
            return results

        dd = {}
        hostlist = []
        for nodecmd in cmds:
            host = nodecmd[0].addr
            if host not in dd:
                dd[host] = []
                hostlist.append(host)
            dd[host].append(nodecmd)

        nodecmdlist = []
        for host in hostlist:
            for zeeknode, cmd, args in dd[host]:
                if helper:
                    cmdargs = [os.path.join(self.config.helperdir, cmd)]
                else:
                    cmdargs = [cmd]

                if shell:
                    if args:
                        cmdargs = ["{} {}".format(cmdargs[0], " ".join(args))]
                else:
                    cmdargs += args

                nodecmdlist.append((zeeknode.addr, cmdargs))
                logging.debug("%s: %s", zeeknode.host, " ".join(cmdargs))

        for host, result in self.sshrunner.exec_multihost_commands(
            nodecmdlist, shell, self.config.commandtimeout
        ):
            nodecmd = dd[host].pop(0)
            zeeknode = nodecmd[0]
            if not isinstance(result, Exception):
                res = result[0]
                out = result[1]
                err = result[2]
                results.append((zeeknode, res == 0, out + err))
                logging.debug("%s: exit code %d", zeeknode.host, res)
            else:
                results.append((zeeknode, False, str(result)))

        return results

    # Run shell commands in parallel on one or more hosts.
    # cmdlines:  a list of the form [ (node, cmdline), ... ]
    #   where "cmdline" is a string to be interpreted by the shell
    #
    # Return value is same as run_cmds.
    def run_shell_cmds(self, cmdlines):
        cmds = [(node, cmdline, []) for node, cmdline in cmdlines]

        return self.run_cmds(cmds, shell=True)

    # A convenience function that calls run_cmds.
    def run_helper(self, cmds, shell=False):
        return self.run_cmds(cmds, shell, True)

    # A convenience function that calls run_cmds.
    # dirs:  a list of the form [ (node, dir), ... ]
    #
    # Returns a list of the form: [ (node, success, output), ... ]
    #   where "success" is a boolean (true if specified directory was created
    #   or already exists).
    def mkdirs(self, dirs):
        results = []
        cmds = []

        for node, dir in dirs:
            cmds += [(node, "mkdir", ["-p", dir])]

        for node, success, output in self.run_cmds(cmds):
            results += [(node, success, output)]

        return results

    # A convenience function that calls run_cmds to remove directories
    # on one or more hosts.
    # dirs:  a list of the form [ (node, dir), ... ]
    #
    # Returns a list of the form: [ (node, success, output), ... ]
    #   where "success" is a boolean (true if specified directory was removed
    #   or does not exist).
    def rmdirs(self, dirs):
        results = []
        cmds = []

        for node, dir in dirs:
            cmds += [(node, f"if [ -d {dir} ]; then rm -rf {dir} ; fi", [])]

        for node, success, output in self.run_cmds(cmds, shell=True):
            results += [(node, success, output)]

        return results

    def host_status(self):
        return self.sshrunner.host_status()



import json
import logging

from ZeekControl import config, version

errmsg = ""

try:
    import broker
    import broker.zeek
except ImportError as e:
    broker = None
    errmsg = e

try:
    websockets_errmsg = None
    import websockets.exceptions as websockets_exceptions
    import websockets.sync.client as websockets_sync_client
    import websockets.version as websockets_version

    # Tested with 11.0 and it seems to work well enough.
    v = websockets_version.version
    if v < "11.0":
        websockets_errmsg = f"Need websockets package version 11.0 or later, have {v}"
except ImportError as e:
    websockets_exceptions = None
    websockets_sync_client = None
    websockets_errmsg = f"Failed to import websockets module ({e!r})"

# Communication with running nodes.


# Sends event to a set of nodes in parallel.
#
# events is a list of tuples of the form (node, event, args, result_event).
#   node:    the destination node.
#   event:   the name of the event to send (node that receiver must subscribe
#            to it as well).
#   args:    a list of event args; each arg must be a data type understood by
#            the Broker module.
#   result_event: name of a event the node sends back. None if no event is
#                 sent back.
#
# Returns a list of tuples (node, success, results_args).
#   If success is True, result_args is a list of arguments as shipped with the
#   result event, or [] if no result_event was specified.
#   If success is False, results_args is a string with an error message.
def send_events_parallel(events, topic):
    clusterbackend = config.Config.clusterbackend
    if config.Config.usewebsocket:
        return ws_send_events(events, topic)
    elif clusterbackend.lower() == "broker":
        return broker_send_events_parallel(events, topic)

    # Error if non-Broker backend in selected and UseWebSocket is not set.
    results = []
    for node, _, _, _ in events:
        results += [
            (
                node,
                False,
                f"command execution with cluster backend '{clusterbackend}' requires UseWebSocket = 1",
            )
        ]
    return results


def broker_send_events_parallel(events, topic):
    results = []
    sent = []

    for node, event, args, result_event in events:
        if not broker:
            results += [(node, False, f"Python bindings for Broker: {errmsg}")]
            continue

        success, endpoint, sub = _send_event_init(
            node, event, args, result_event, topic
        )

        if success and result_event:
            sent += [(node, result_event, endpoint, sub)]
        else:
            sub.reset()
            endpoint.shutdown()
            results += [(node, success, "")]

    for node, result_event, endpoint, sub in sent:
        success, result_args = _send_event_wait(node, result_event, endpoint, sub)
        sub.reset()
        endpoint.shutdown()
        results += [(node, success, result_args)]

    return results


def _send_event_init(node, event, args, result_event, topic):
    host = node.addr
    endpoint = broker.Endpoint()
    subscriber = endpoint.make_subscriber(topic)

    with endpoint.make_status_subscriber(True) as status_subscriber:
        endpoint.peer(host, node.getPort(), 1)

        tries = 0

        while True:
            msgs = status_subscriber.get(1, 1)

            for msg in msgs:
                if isinstance(msg, broker.Status):
                    if msg.code() == broker.SC.PeerAdded:
                        ev = broker.zeek.Event(event, *args)
                        endpoint.publish(topic + "/" + repr(msg.context()), ev)
                        logging.debug(
                            "broker: %s(%s) to node %s",
                            event,
                            ", ".join(args),
                            node.name,
                        )
                        return (True, endpoint, subscriber)

            tries += 1

            if tries > config.Config.commtimeout:
                return (False, "time-out", None)


def _send_event_wait(node, result_event, bc, sub):
    if not result_event:
        return (True, [])

    # Wait for reply event.
    tries = 0

    while True:
        msgs = sub.get(1, 1)

        for msg in msgs:
            (topic, event) = msg
            ev = broker.zeek.Event(event)
            args = ev.args()
            logging.debug(
                "broker: %s(%s) from node %s", result_event, ", ".join(args), node.name
            )
            return (True, args)

        tries += 1

        if tries > config.Config.commtimeout:
            logging.debug("broker: timeout during receive from node %s", node.name)
            return (False, "time-out")


class WebSocketError(Exception):
    pass


class WebSocketClient:
    """
    Small wrapper around the websockets library to send
    and receive simple Zeek events via v1/messages/json.

    This is good enough for Zeekctl. If you have a use for
    this, put it in a library :-)
    """

    DEFAULT_TIMEOUT = 10.0

    def __init__(self, c, *, timeout=DEFAULT_TIMEOUT):
        self.__c = c
        self.__timeout = timeout

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__c.__exit__(exc_type, exc_value, traceback)

    def send_json(self, data):
        try:
            self.__c.send(json.dumps(data))
        except (
            websockets_exceptions.ConnectionClosed,
            websockets_exceptions.ConnectionClosedError,
        ) as e:
            raise WebSocketError(e)

    def recv_json(self):
        try:
            return json.loads(self.__c.recv(timeout=self.__timeout))
        except (
            TimeoutError,
            websockets_exceptions.ConnectionClosed,
            websockets_exceptions.ConnectionClosedError,
        ) as e:
            raise WebSocketError(e)

    def v1_hello(self, subscriptions):
        """
        Send the initial subscription array for v1/messages/json.
        """
        self.send_json(subscriptions)
        ack = self.recv_json()
        if ack.get("type") != "ack":
            raise WebSocketError(f"Unexpected ack type {ack!r}")

        return ack

    def v1_event(self, topic, name, args=None):
        """
        Send an event in v1/messages/json format.
        """
        ev_args = [self.v1_arg(a) for a in (args or [])]
        d = {
            "type": "data-message",
            "topic": topic,
            "@data-type": "vector",
            "data": [
                {"@data-type": "count", "data": 1},  # Format
                {"@data-type": "count", "data": 1},  # Type
                {
                    "@data-type": "vector",
                    "data": [  # Event vector
                        {"@data-type": "string", "data": name},
                        {"@data-type": "vector", "data": ev_args},
                    ],
                },
            ],
        }

        return self.send_json(d)

    def v1_recv_event(self):
        """
        Wait for a JSON message and interpret it as v1/messages/json event.

        Returns a (topic, name, args) tuple.
        """
        d = self.recv_json()
        if d["type"] != "data-message":
            raise WebSocketError(f"unexpected event reply {d!r}")

        # See event_v1() for the format.
        ev_name = d["data"][2]["data"][0]["data"]
        ev_args = d["data"][2]["data"][1]["data"]
        ev_args_vals = [ev_arg["data"] for ev_arg in ev_args]
        return d["topic"], ev_name, ev_args_vals

    @staticmethod
    def v1_arg(arg):
        data_type = None
        data = None

        if isinstance(arg, int):
            data_type = "count"
            data = arg
        elif isinstance(arg, str):
            data_type = "string"
            data = arg
        else:
            raise TypeError(f"Unsupported arg {arg!r} of type {type(arg)}")

        return {
            "@data-type": data_type,
            "data": data,
        }

    @staticmethod
    def connect(url, *, application_name=None, timeout=None):
        additional_headers = {}
        if application_name:
            additional_headers["X-Application-Name"] = application_name

        try:
            return WebSocketClient(
                websockets_sync_client.connect(
                    uri=url,
                    additional_headers=additional_headers,
                    open_timeout=timeout,
                    close_timeout=timeout,
                ),
                timeout=timeout,
            )
        except (ConnectionRefusedError, TimeoutError) as e:
            raise WebSocketError(e)


def ws_send_events(events, topic):
    """
    Use a short-lived WebSocket connection to the manager and send
    events serially to individual node topics and await the response.

    This will not work correctly if two zeekctl processes do this at
    the same time. If we want this, we'd need to make the topic unique
    and pass it as event parameter so the control scripts can use it
    for their reply. Punt on that for now, I think the same issue
    exists with the native Broker integration, too.
    """
    results = []

    if websockets_errmsg:
        for node, event, args, result_event in events:
            results += [(node, False, websockets_errmsg)]
        return results

    if config.Config.websocketurl:
        url = config.Config.websocketurl
    else:
        host = config.Config.websockethost
        port = config.Config.websocketport
        url = f"ws://{host}:{port}/v1/messages/json"

    try:
        ws = WebSocketClient.connect(
            url=url,
            application_name=f"zeekctl/{version.VERSION}",
            timeout=config.Config.websockettimeout,
        )
    except WebSocketError as e:
        for node, event, args, result_event in events:
            results += [(node, False, str(e))]
        return results

    with ws:
        # Subscribe to the zeek/control reply topic.
        try:
            ws.v1_hello([topic])
        except WebSocketError as e:
            for node, event, args, result_event in events:
                results += [(node, False, str(e))]
            return results

        for node, event, args, result_event in events:
            # Use the topic separator configured by the backend for publishing
            # to individual node topics.
            topic_sep = config.Config.clustertopicseparator
            topic = topic_sep.join(["zeek", "cluster", "node", node.name, ""])

            try:
                ws.v1_event(topic, event, args)
                rtopic, rname, rargs = ws.v1_recv_event()
            except WebSocketError as e:
                results += [(node, False, repr(e))]
                continue
            else:
                # Did we even receive the right event?
                if result_event != rname:
                    results += [
                        (node, False, f"expected '{result_event}' got '{rname}'")
                    ]
                    continue

                # Figure out which node sent the reply. It's the last part
                # in the reply topic. The / is hard-coded in the control
                # scripts, so we try that first if it's found, else fallback
                # to topic_sep (e.g, could be "." for ZeroMQ, NATS or RabbitMQ)
                if "/" in rtopic:
                    rnode = rtopic.rsplit("/", 1)[-1]
                else:
                    rnode = rtopic.rsplit(topic_sep, 1)[-1]

                # Expect a reply from the addressed node, or an empty node if
                # this a standalone setup.
                if rnode != node.name and not (
                    config.Config.standalone and rnode == ""
                ):
                    results += [
                        (
                            node,
                            False,
                            f"'unexpected {rnode}' in '{topic}', expected '{node.name}'",
                        )
                    ]
                    continue

                results += [(node, True, rargs)]

    return results
